"""API response object."""

from __future__ import annotations

from collections.abc import Mapping
from typing import Generic
from typing import TypeVar

from pydantic import BaseModel
from pydantic import Field
from pydantic import StrictBytes
from pydantic import StrictInt

T = TypeVar("T")


class ApiResponse(BaseModel, Generic[T]):
    """API response object"""

    status_code: StrictInt = Field(description="HTTP status code")
    headers: Mapping[str, str] | None = Field(None, description="HTTP headers")
    data: T = Field(description="Deserialized data given the data type")
    raw_data: StrictBytes = Field(description="Raw data (HTTP response body)")

    model_config = {
        "arbitrary_types_allowed": True,
    }



"""Handle reading, updating and writing of metadata."""

from __future__ import annotations

import copy
import json
import logging
import warnings
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import TYPE_CHECKING
from typing import cast

import datadoc_model.all_optional.model as all_optional_model
import datadoc_model.required.model as required_model
from datadoc_model.all_optional.model import DataSetStatus

from dapla_metadata._shared import config
from dapla_metadata.dapla import user_info
from dapla_metadata.datasets.dapla_dataset_path_info import DaplaDatasetPathInfo
from dapla_metadata.datasets.dataset_parser import DatasetParser
from dapla_metadata.datasets.model_backwards_compatibility import (
    is_metadata_in_container_structure,
)
from dapla_metadata.datasets.model_backwards_compatibility import upgrade_metadata
from dapla_metadata.datasets.model_validation import ValidateDatadocMetadata
from dapla_metadata.datasets.statistic_subject_mapping import StatisticSubjectMapping
from dapla_metadata.datasets.utility.constants import (
    DEFAULT_SPATIAL_COVERAGE_DESCRIPTION,
)
from dapla_metadata.datasets.utility.constants import INCONSISTENCIES_MESSAGE
from dapla_metadata.datasets.utility.constants import METADATA_DOCUMENT_FILE_SUFFIX
from dapla_metadata.datasets.utility.constants import NUM_OBLIGATORY_DATASET_FIELDS
from dapla_metadata.datasets.utility.constants import NUM_OBLIGATORY_VARIABLES_FIELDS
from dapla_metadata.datasets.utility.utils import OptionalDatadocMetadataType
from dapla_metadata.datasets.utility.utils import calculate_percentage
from dapla_metadata.datasets.utility.utils import derive_assessment_from_state
from dapla_metadata.datasets.utility.utils import get_timestamp_now
from dapla_metadata.datasets.utility.utils import merge_variables
from dapla_metadata.datasets.utility.utils import normalize_path
from dapla_metadata.datasets.utility.utils import (
    num_obligatory_dataset_fields_completed,
)
from dapla_metadata.datasets.utility.utils import (
    num_obligatory_variables_fields_completed,
)
from dapla_metadata.datasets.utility.utils import override_dataset_fields
from dapla_metadata.datasets.utility.utils import set_dataset_owner
from dapla_metadata.datasets.utility.utils import set_default_values_dataset
from dapla_metadata.datasets.utility.utils import set_default_values_variables

if TYPE_CHECKING:
    import pathlib
    from datetime import datetime

    from cloudpathlib import CloudPath


logger = logging.getLogger(__name__)


class InconsistentDatasetsWarning(UserWarning):
    """Existing and new datasets differ significantly from one another."""


class InconsistentDatasetsError(ValueError):
    """Existing and new datasets differ significantly from one another."""


class Datadoc:
    """Handle reading, updating and writing of metadata.

    If a metadata document exists, it is this information that is loaded. Nothing
    is inferred from the dataset. If only a dataset path is supplied the metadata
    document path is build based on the dataset path.

    Example: /path/to/dataset.parquet -> /path/to/dataset__DOC.json

    Attributes:
        dataset_path: A file path to the path to where the dataset is stored.
        metadata_document_path: A path to a metadata document if it exists.
        statistic_subject_mapping: An instance of StatisticSubjectMapping.
    """

    def __init__(
        self,
        dataset_path: str | None = None,
        metadata_document_path: str | None = None,
        statistic_subject_mapping: StatisticSubjectMapping | None = None,
        errors_as_warnings: bool = False,
        validate_required_fields_on_existing_metadata: bool = False,
    ) -> None:
        """Initialize the Datadoc instance.

        If a dataset path is supplied, it attempts to locate and load the
        corresponding metadata document. If no dataset path is provided, the class
        is instantiated without loading any metadata.

        Args:
            dataset_path: The file path to the dataset. Defaults to None.
            metadata_document_path: The file path to the metadata document.
                Defaults to None.
            statistic_subject_mapping: An instance of StatisticSubjectMapping.
                Defaults to None
            errors_as_warnings: Disable raising exceptions if inconsistencies
                are found between existing and extracted metadata.
            validate_required_fields_on_existing_metadata: Use a Pydantic model
                which validates whether required fields are present when reading
                in an existing metadata file.
        """
        self._statistic_subject_mapping = statistic_subject_mapping
        self.errors_as_warnings = errors_as_warnings
        self.validate_required_fields_on_existing_metadata = (
            validate_required_fields_on_existing_metadata
        )
        self.metadata_document: pathlib.Path | CloudPath | None = None
        self.container: all_optional_model.MetadataContainer | None = None
        self.dataset_path: pathlib.Path | CloudPath | None = None
        self.dataset = all_optional_model.Dataset()
        self.variables: list = []
        self.variables_lookup: dict[str, all_optional_model.Variable] = {}
        self.explicitly_defined_metadata_document = False
        self.dataset_consistency_status: list = []
        if metadata_document_path:
            self.metadata_document = normalize_path(metadata_document_path)
            self.explicitly_defined_metadata_document = True
            if not self.metadata_document.exists():
                msg = f"Metadata document does not exist! Provided path: {self.metadata_document}"
                raise ValueError(
                    msg,
                )
        if dataset_path:
            self.dataset_path = normalize_path(dataset_path)
            if not metadata_document_path:
                self.metadata_document = self.build_metadata_document_path(
                    self.dataset_path,
                )
        if metadata_document_path or dataset_path:
            self._extract_metadata_from_files()

    def _extract_metadata_from_files(self) -> None:
        """Read metadata from an existing metadata document or create one.

        If a metadata document exists, it reads and extracts metadata from it.
        If no metadata document is found, it creates metadata from scratch by
        extracting information from the dataset file.

        This method ensures that:
        - Metadata is extracted from an existing document if available.
        - If metadata is not available, it is extracted from the dataset file.
        - The dataset ID is set if not already present.
        - Default values are set for variables, particularly the variable role on
            creation.
        - Default values for variables ID and 'is_personal_data' are set if the
            values are None.
        - The 'contains_personal_data' attribute is set to False if not specified.
        - A lookup dictionary for variables is created based on their short names.
        """
        extracted_metadata: all_optional_model.DatadocMetadata | None = None
        existing_metadata: OptionalDatadocMetadataType = None

        if self.metadata_document and self.metadata_document.exists():
            existing_metadata = self._extract_metadata_from_existing_document(
                self.metadata_document,
            )

        if (
            self.dataset_path is not None
            and self.dataset == all_optional_model.Dataset()
            and len(self.variables) == 0
        ):
            extracted_metadata = self._extract_metadata_from_dataset(self.dataset_path)

        if extracted_metadata is not None:
            existing_file_path = self._get_existing_file_path(extracted_metadata)
            if (
                self.dataset_path
                and existing_file_path is not None
                and extracted_metadata is not None
                and existing_metadata is not None
            ):
                self.dataset_consistency_status = self._check_dataset_consistency(
                    self.dataset_path,
                    Path(existing_file_path),
                    extracted_metadata,
                    existing_metadata,
                )

        if (
            self.dataset_path
            and self.explicitly_defined_metadata_document
            and self.metadata_document is not None
            and self.metadata_document.exists()
            and extracted_metadata is not None
            and existing_metadata is not None
        ):
            self._check_ready_to_merge(
                self.dataset_consistency_status,
                errors_as_warnings=self.errors_as_warnings,
            )
            merged_metadata = self._merge_metadata(
                extracted_metadata,
                existing_metadata,
            )
            # We need to override this so that the document gets saved to the correct
            # location, otherwise we would overwrite the existing document!
            self.metadata_document = self.build_metadata_document_path(
                self.dataset_path,
            )
            self._set_metadata(merged_metadata)
        else:
            self._set_metadata(existing_metadata or extracted_metadata)

        set_default_values_variables(self.variables)
        set_default_values_dataset(cast("all_optional_model.Dataset", self.dataset))
        set_dataset_owner(self.dataset)
        self._create_variables_lookup()

    def _get_existing_file_path(
        self,
        extracted_metadata: all_optional_model.DatadocMetadata | None,
    ) -> str:
        if (
            extracted_metadata is not None
            and extracted_metadata.dataset is not None
            and extracted_metadata.dataset.file_path is not None
        ):
            return extracted_metadata.dataset.file_path
        msg = "Could not access existing dataset file path"
        raise ValueError(msg)

    def _set_metadata(
        self,
        merged_metadata: OptionalDatadocMetadataType,
    ) -> None:
        if not merged_metadata or not (
            merged_metadata.dataset and merged_metadata.variables
        ):
            msg = "Could not read metadata"
            raise ValueError(msg)
        self.dataset = cast("all_optional_model.Dataset", merged_metadata.dataset)
        self.variables = merged_metadata.variables

    def _create_variables_lookup(self) -> None:
        self.variables_lookup = {
            v.short_name: v for v in self.variables if v.short_name
        }

    @staticmethod
    def _check_dataset_consistency(
        new_dataset_path: Path | CloudPath,
        existing_dataset_path: Path,
        extracted_metadata: all_optional_model.DatadocMetadata,
        existing_metadata: OptionalDatadocMetadataType,
    ) -> list[dict[str, object]]:
        """Run consistency tests.

        Args:
            new_dataset_path: Path to the dataset to be documented.
            existing_dataset_path: Path stored in the existing metadata.
            extracted_metadata: Metadata extracted from a physical dataset.
            existing_metadata: Metadata from a previously created metadata document.

        Returns:
            List if dict with property name and boolean success flag
        """
        new_dataset_path_info = DaplaDatasetPathInfo(new_dataset_path)
        existing_dataset_path_info = DaplaDatasetPathInfo(existing_dataset_path)
        return [
            {
                "name": "Bucket name",
                "success": (
                    new_dataset_path_info.bucket_name
                    == existing_dataset_path_info.bucket_name
                ),
            },
            {
                "name": "Data product name",
                "success": (
                    new_dataset_path_info.statistic_short_name
                    == existing_dataset_path_info.statistic_short_name
                ),
            },
            {
                "name": "Dataset state",
                "success": (
                    new_dataset_path_info.dataset_state
                    == existing_dataset_path_info.dataset_state
                ),
            },
            {
                "name": "Dataset short name",
                "success": (
                    new_dataset_path_info.dataset_short_name
                    == existing_dataset_path_info.dataset_short_name
                ),
            },
            {
                "name": "Variable names",
                "success": (
                    existing_metadata is not None
                    and {v.short_name for v in extracted_metadata.variables or []}
                    == {v.short_name for v in existing_metadata.variables or []}
                ),
            },
            {
                "name": "Variable datatypes",
                "success": (
                    existing_metadata is not None
                    and [v.data_type for v in extracted_metadata.variables or []]
                    == [v.data_type for v in existing_metadata.variables or []]
                ),
            },
        ]

    @staticmethod
    def _check_ready_to_merge(
        results: list[dict[str, object]], *, errors_as_warnings: bool
    ) -> None:
        """Check if the datasets are consistent enough to make a successful merge of metadata.

        Args:
            results: List if dict with property name and boolean success flag
            errors_as_warnings: True if failing checks should be raised as warnings, not errors.

        Raises:
            InconsistentDatasetsError: If inconsistencies are found and `errors_as_warnings == False`
        """
        if failures := [result for result in results if not result["success"]]:
            msg = f"{INCONSISTENCIES_MESSAGE} {', '.join(str(f['name']) for f in failures)}"
            if errors_as_warnings:
                warnings.warn(
                    message=msg,
                    category=InconsistentDatasetsWarning,
                    stacklevel=2,
                )
            else:
                raise InconsistentDatasetsError(
                    msg,
                )

    @staticmethod
    def _merge_metadata(
        extracted_metadata: all_optional_model.DatadocMetadata | None,
        existing_metadata: OptionalDatadocMetadataType,
    ) -> all_optional_model.DatadocMetadata:
        if not existing_metadata:
            logger.warning(
                "No existing metadata found, no merge to perform. Continuing with extracted metadata.",
            )
            return extracted_metadata or all_optional_model.DatadocMetadata()

        if not extracted_metadata:
            return cast("all_optional_model.DatadocMetadata", existing_metadata)

        # Use the extracted metadata as a base
        merged_metadata = all_optional_model.DatadocMetadata(
            dataset=copy.deepcopy(extracted_metadata.dataset),
            variables=[],
        )

        override_dataset_fields(
            merged_metadata=merged_metadata,
            existing_metadata=cast(
                "all_optional_model.DatadocMetadata", existing_metadata
            ),
        )

        # Merge variables.
        # For each extracted variable, copy existing metadata into the merged metadata
        return merge_variables(
            existing_metadata=existing_metadata,
            extracted_metadata=extracted_metadata,
            merged_metadata=merged_metadata,
        )

    def _extract_metadata_from_existing_document(
        self,
        document: pathlib.Path | CloudPath,
    ) -> OptionalDatadocMetadataType:
        """Read metadata from an existing metadata document.

        If an existing metadata document is available, this method reads and
        loads the metadata from it. It validates and upgrades the metadata as
        necessary. If we have read in a file with an empty "datadoc" structure
        the process ends.
        A typical example causing a empty datadoc is a file produced from a
        pseudonymization process.

        Args:
            document: A path to the existing metadata document.

        Raises:
            json.JSONDecodeError: If the metadata document cannot be parsed.
            pydantic.ValidationError: If the data does not successfully validate.
        """
        metadata_model = (
            required_model
            if self.validate_required_fields_on_existing_metadata
            else all_optional_model
        )
        fresh_metadata = {}
        try:
            with document.open(mode="r", encoding="utf-8") as file:
                fresh_metadata = json.load(file)
            logger.info("Opened existing metadata file %s", document)
            fresh_metadata = upgrade_metadata(
                fresh_metadata,
            )
            if is_metadata_in_container_structure(fresh_metadata):
                self.container = metadata_model.MetadataContainer.model_validate_json(
                    json.dumps(fresh_metadata),
                )
                datadoc_metadata = fresh_metadata["datadoc"]
            else:
                datadoc_metadata = fresh_metadata
            if datadoc_metadata is None:
                return None
            return metadata_model.DatadocMetadata.model_validate_json(
                json.dumps(datadoc_metadata),
            )
        except json.JSONDecodeError:
            logger.warning(
                "Could not open existing metadata file %s. \
                    Falling back to collecting data from the dataset",
                document,
                exc_info=True,
            )
            return None

    def _extract_subject_field_from_path(
        self,
        dapla_dataset_path_info: DaplaDatasetPathInfo,
    ) -> str | None:
        """Extract the statistic short name from the dataset file path.

        Map the extracted statistic short name to its corresponding statistical
        subject.

        Args:
            dapla_dataset_path_info: The object representing the decomposed file
                path.

        Returns:
            The code for the statistical subject or None if we couldn't map to one.
        """
        if self._statistic_subject_mapping is None:
            with ThreadPoolExecutor(max_workers=12) as executor:
                return StatisticSubjectMapping(
                    executor,
                    config.get_statistical_subject_source_url(),
                ).get_secondary_subject(
                    dapla_dataset_path_info.statistic_short_name,
                )
        else:
            return self._statistic_subject_mapping.get_secondary_subject(
                dapla_dataset_path_info.statistic_short_name,
            )

    def _extract_metadata_from_dataset(
        self,
        dataset: pathlib.Path | CloudPath,
    ) -> all_optional_model.DatadocMetadata:
        """Obtain what metadata we can from the dataset itself.

        This makes it easier for the user by 'pre-filling' certain fields.
        Certain elements are dependent on the dataset being saved according
        to SSB's standard.

        Args:
            dataset: The path to the dataset file, which can be a local or
                cloud path.

        Side Effects:
            Updates the following instance attributes:
                - ds_schema: An instance of DatasetParser initialized for the
                    given dataset file.
                - dataset: An instance of model.Dataset with pre-filled metadata
                    fields.
                - variables: A list of fields extracted from the dataset schema.
        """
        dapla_dataset_path_info = DaplaDatasetPathInfo(dataset)
        metadata = all_optional_model.DatadocMetadata()

        metadata.dataset = all_optional_model.Dataset(
            short_name=dapla_dataset_path_info.dataset_short_name,
            dataset_state=dapla_dataset_path_info.dataset_state,
            dataset_status=DataSetStatus.DRAFT,
            assessment=(
                derive_assessment_from_state(
                    dapla_dataset_path_info.dataset_state,
                )
                if dapla_dataset_path_info.dataset_state is not None
                else None
            ),
            version=dapla_dataset_path_info.dataset_version,
            contains_data_from=dapla_dataset_path_info.contains_data_from,
            contains_data_until=dapla_dataset_path_info.contains_data_until,
            file_path=str(self.dataset_path),
            metadata_created_by=user_info.get_user_info_for_current_platform().short_email,
            subject_field=self._extract_subject_field_from_path(
                dapla_dataset_path_info,
            ),
            spatial_coverage_description=DEFAULT_SPATIAL_COVERAGE_DESCRIPTION,
        )
        metadata.variables = DatasetParser.for_file(dataset).get_fields()
        return metadata

    @staticmethod
    def build_metadata_document_path(
        dataset_path: pathlib.Path | CloudPath,
    ) -> pathlib.Path | CloudPath:
        """Build the path to the metadata document corresponding to the given dataset.

        Args:
            dataset_path: Path to the dataset we wish to create metadata for.
        """
        return dataset_path.parent / (dataset_path.stem + METADATA_DOCUMENT_FILE_SUFFIX)

    def datadoc_model(self) -> all_optional_model.MetadataContainer:
        """Return the underlying datadoc model."""
        datadoc: ValidateDatadocMetadata = ValidateDatadocMetadata(
            percentage_complete=self.percent_complete,
            dataset=self.dataset,
            variables=self.variables,
        )
        if self.container:
            res = copy.deepcopy(self.container)
            res.datadoc = datadoc
            return res
        return all_optional_model.MetadataContainer(datadoc=datadoc)

    def write_metadata_document(self) -> None:
        """Write all currently known metadata to file.

        Side Effects:
            - Updates the dataset's metadata_last_updated_date and
                metadata_last_updated_by attributes.
            - Updates the dataset's file_path attribute.
            - Validates the metadata model and stores it in a MetadataContainer.
            - Writes the validated metadata to a file if the metadata_document
                attribute is set.
            - Logs the action and the content of the metadata document.

        Raises:
            ValueError: If no metadata document is specified for saving.
        """
        timestamp: datetime = get_timestamp_now()
        self.dataset.metadata_last_updated_date = timestamp
        self.dataset.metadata_last_updated_by = (
            user_info.get_user_info_for_current_platform().short_email
        )
        self.dataset.file_path = str(self.dataset_path)
        datadoc: ValidateDatadocMetadata = ValidateDatadocMetadata(
            percentage_complete=self.percent_complete,
            dataset=self.dataset,
            variables=self.variables,
        )
        if self.container:
            self.container.datadoc = datadoc
        else:
            self.container = all_optional_model.MetadataContainer(datadoc=datadoc)
        if self.metadata_document:
            content = self.container.model_dump_json(indent=4)
            self.metadata_document.write_text(content)
            logger.info("Saved metadata document %s", self.metadata_document)
            logger.info(
                "Metadata content",
                extra={"metadata_content": json.loads(content)},
            )
        else:
            msg = "No metadata document to save"
            raise ValueError(msg)

    @property
    def percent_complete(self) -> int:
        """The percentage of obligatory metadata completed.

        A metadata field is counted as complete when any non-None value is
        assigned. Used for a live progress bar in the UI, as well as being
        saved in the datadoc as a simple quality indicator.
        """
        num_all_fields = NUM_OBLIGATORY_DATASET_FIELDS + (
            NUM_OBLIGATORY_VARIABLES_FIELDS * len(self.variables)
        )
        num_set_fields = num_obligatory_dataset_fields_completed(
            self.dataset,
        ) + num_obligatory_variables_fields_completed(self.variables)
        return calculate_percentage(num_set_fields, num_all_fields)

    def add_pseudonymization(
        self,
        variable_short_name: str,
        pseudonymization: all_optional_model.Pseudonymization | None = None,
    ) -> None:
        """Adds a new pseudo variable to the list of pseudonymized variables also sets is_personal_data to true.

        If there is no pseudonymization supplied an empty Pseudonymization structure will be added to the model.

        Args:
            variable_short_name: The short name for the variable that one wants to update the pseudo for.
            pseudonymization: The updated pseudonymization.

        """
        variable = self.variables_lookup[variable_short_name]
        variable.pseudonymization = (
            pseudonymization or all_optional_model.Pseudonymization()
        )
        variable.is_personal_data = True

    def remove_pseudonymization(self, variable_short_name: str) -> None:
        """Removes a pseudo variable by using the shortname.

        Updates the pseudo variable lookup by creating a new one.
        Sets is_personal_data to non pseudonymized encrypted personal data.

        Args:
            variable_short_name: The short name for the variable that one wants to remove the pseudo for.
        """
        if self.variables_lookup[variable_short_name].pseudonymization is not None:
            self.variables_lookup[variable_short_name].pseudonymization = None
            self.variables_lookup[variable_short_name].is_personal_data = False



"""Internal Variable Definitions Administration API

## Introduction  Variable Definitions are centralized definitions of concrete variables which are typically present in multiple datasets. Variable Definitions support standardization of data and metadata and facilitate sharing and joining of data by clarifying when variables have an identical definition.  ## Maintenance of Variable Definitions This API allows for creation, maintenance and access of Variable Definitions.  ### Ownership Creation and maintenance of variables may only be performed by Statistics Norway employees representing a specific Dapla team, who are defined as the owners of a given Variable Definition. The team an owner represents must be specified when making a request through the `active_group` query parameter. All maintenance is to be performed by the owners, with no intervention from administrators.  ### Status All Variable Definitions have an associated status. The possible values for status are `DRAFT`, `PUBLISHED_INTERNAL` and `PUBLISHED_EXTERNAL`.  #### Draft When a Variable Definition is created it is assigned the status `DRAFT`. Under this status the Variable Definition is:  - Only visible to Statistics Norway employees. - Mutable (it may be changed directly without need for versioning). - Not suitable to refer to from other systems.  This status may be changed to `PUBLISHED_INTERNAL` or `PUBLISHED_EXTERNAL` with a direct update.  #### Published Internal Under this status the Variable Definition is:  - Only visible to Statistics Norway employees. - Immutable (all changes are versioned). - Suitable to refer to in internal systems for statistics production. - Not suitable to refer to for external use (for example in Statistikkbanken).  This status may be changed to `PUBLISHED_EXTERNAL` by creating a Patch version.  #### Published External Under this status the Variable Definition is:  - Visible to the general public. - Immutable (all changes are versioned). - Suitable to refer to from any system.  This status may not be changed as it would break immutability. If a Variable Definition is no longer relevant then its period of validity should be ended by specifying a `valid_until` date in a Patch version.  ### Immutability Variable Definitions are immutable. This means that any changes must be performed in a strict versioning system. Consumers can avoid being exposed to breaking changes by specifying a `date_of_validity` when they request a Variable Definition.  #### Patches Patches are for changes which do not affect the fundamental meaning of the Variable Definition.  #### Validity Periods Validity Periods are versions with a period defined by a `valid_from` date and optionally a `valid_until` date. If the fundamental meaning of a Variable Definition is to be changed, it should be done by creating a new Validity Period.

The version of the OpenAPI document: 0.1
Contact: metadata@ssb.no
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401
from datetime import date
from typing import Annotated
from typing import Any
from typing import ClassVar

from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic import Field
from pydantic import StrictBool
from pydantic import StrictStr
from typing_extensions import Self

from ..models.contact import Contact
from ..models.language_string_type import LanguageStringType


class ValidityPeriod(BaseModel):
    """Create a new Validity Period on a Published Variable Definition."""

    name: LanguageStringType | None = None
    definition: LanguageStringType
    classification_reference: StrictStr | None = Field(
        default=None,
        description="ID of a classification or code list from Klass. The given classification defines all possible values for the defined variable.",
    )
    unit_types: list[Annotated[str, Field(min_length=1, strict=True)]] | None = Field(
        default=None,
        description="A list of one or more unit types, e.g. person, vehicle, household. Must be defined as codes from https://www.ssb.no/klass/klassifikasjoner/702.",
    )
    subject_fields: list[Annotated[str, Field(min_length=1, strict=True)]] | None = (
        Field(
            default=None,
            description="A list of subject fields that the variable is used in. Must be defined as codes from https://www.ssb.no/klass/klassifikasjoner/618.",
        )
    )
    contains_special_categories_of_personal_data: StrictBool | None = Field(
        default=None,
        description="True if variable instances contain particularly sensitive information. Applies even if the information or identifiers are pseudonymized. Information within the following categories are regarded as particularly sensitive: Ethnicity, Political alignment, Religion, Philosophical beliefs, Union membership, Genetics, Biometrics, Health, Sexual relations, Sexual orientation",
    )
    measurement_type: StrictStr | None = Field(
        default=None,
        description="Type of measurement for the variable, e.g. length, volume, currency. Must be defined as codes from https://www.ssb.no/klass/klassifikasjoner/303",
    )
    valid_from: date = Field(
        description="The variable definition is valid from this date inclusive"
    )
    external_reference_uri: StrictStr | None = Field(
        default=None, description="A link (URI) to an external definition/documentation"
    )
    comment: LanguageStringType | None = None
    related_variable_definition_uris: list[StrictStr] | None = Field(
        default=None,
        description="Link(s) to related definitions of variables - a list of one or more definitions. For example for a variable after-tax income it could be relevant to link to definitions of income from work, property income etc.",
    )
    contact: Contact | None = None
    __properties: ClassVar[list[str]] = [
        "name",
        "definition",
        "classification_reference",
        "unit_types",
        "subject_fields",
        "contains_special_categories_of_personal_data",
        "measurement_type",
        "valid_from",
        "external_reference_uri",
        "comment",
        "related_variable_definition_uris",
        "contact",
    ]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Self | None:
        """Create an instance of ValidityPeriod from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: set[str] = set([])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of name
        if self.name:
            _dict["name"] = self.name.to_dict()
        # override the default output from pydantic by calling `to_dict()` of definition
        if self.definition:
            _dict["definition"] = self.definition.to_dict()
        # override the default output from pydantic by calling `to_dict()` of comment
        if self.comment:
            _dict["comment"] = self.comment.to_dict()
        # override the default output from pydantic by calling `to_dict()` of contact
        if self.contact:
            _dict["contact"] = self.contact.to_dict()
        # set to None if classification_reference (nullable) is None
        # and model_fields_set contains the field
        if (
            self.classification_reference is None
            and "classification_reference" in self.model_fields_set
        ):
            _dict["classification_reference"] = None

        # set to None if unit_types (nullable) is None
        # and model_fields_set contains the field
        if self.unit_types is None and "unit_types" in self.model_fields_set:
            _dict["unit_types"] = None

        # set to None if subject_fields (nullable) is None
        # and model_fields_set contains the field
        if self.subject_fields is None and "subject_fields" in self.model_fields_set:
            _dict["subject_fields"] = None

        # set to None if contains_special_categories_of_personal_data (nullable) is None
        # and model_fields_set contains the field
        if (
            self.contains_special_categories_of_personal_data is None
            and "contains_special_categories_of_personal_data" in self.model_fields_set
        ):
            _dict["contains_special_categories_of_personal_data"] = None

        # set to None if measurement_type (nullable) is None
        # and model_fields_set contains the field
        if (
            self.measurement_type is None
            and "measurement_type" in self.model_fields_set
        ):
            _dict["measurement_type"] = None

        # set to None if external_reference_uri (nullable) is None
        # and model_fields_set contains the field
        if (
            self.external_reference_uri is None
            and "external_reference_uri" in self.model_fields_set
        ):
            _dict["external_reference_uri"] = None

        # set to None if related_variable_definition_uris (nullable) is None
        # and model_fields_set contains the field
        if (
            self.related_variable_definition_uris is None
            and "related_variable_definition_uris" in self.model_fields_set
        ):
            _dict["related_variable_definition_uris"] = None

        return _dict

    @classmethod
    def from_dict(cls, obj: dict[str, Any] | None) -> Self | None:
        """Create an instance of ValidityPeriod from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate(
            {
                "name": LanguageStringType.from_dict(obj["name"])
                if obj.get("name") is not None
                else None,
                "definition": LanguageStringType.from_dict(obj["definition"])
                if obj.get("definition") is not None
                else None,
                "classification_reference": obj.get("classification_reference"),
                "unit_types": obj.get("unit_types"),
                "subject_fields": obj.get("subject_fields"),
                "contains_special_categories_of_personal_data": obj.get(
                    "contains_special_categories_of_personal_data"
                ),
                "measurement_type": obj.get("measurement_type"),
                "valid_from": obj.get("valid_from"),
                "external_reference_uri": obj.get("external_reference_uri"),
                "comment": LanguageStringType.from_dict(obj["comment"])
                if obj.get("comment") is not None
                else None,
                "related_variable_definition_uris": obj.get(
                    "related_variable_definition_uris"
                ),
                "contact": Contact.from_dict(obj["contact"])
                if obj.get("contact") is not None
                else None,
            }
        )
        return _obj



from dapla_metadata.variable_definitions._generated.vardef_client.api_client import (
    ApiClient,
)
from dapla_metadata.variable_definitions._generated.vardef_client.configuration import (
    Configuration,
)
from dapla_metadata.variable_definitions._utils.config import (
    get_vardef_client_configuration,
)


class VardefClient:
    _client: ApiClient | None = None
    _config: Configuration | None = None

    @classmethod
    def get_config(cls) -> Configuration | None:
        """Get the client configuration object."""
        return cls._config

    @classmethod
    def set_config(cls, config: Configuration) -> None:
        """Set the client configuration object."""
        cls._config = config

    @classmethod
    def get_client(cls) -> ApiClient:
        if not cls._client:
            if not cls._config:
                cls._config = get_vardef_client_configuration()
            cls._client = ApiClient(cls._config)
        return cls._client



"""Variable Definitions

## Introduction  Variable Definitions are centralized definitions of concrete variables which are typically present in multiple datasets. Variable Definitions support standardization of data and metadata and facilitate sharing and joining of data by clarifying when variables have an identical definition.  ## Maintenance of Variable Definitions This API allows for creation, maintenance and access of Variable Definitions.  ### Ownership Creation and maintenance of variables may only be performed by Statistics Norway employees representing a specific Dapla team, who are defined as the owners of a given Variable Definition. The team an owner represents must be specified when making a request through the `active_group` query parameter. All maintenance is to be performed by the owners, with no intervention from administrators.  ### Status All Variable Definitions have an associated status. The possible values for status are `DRAFT`, `PUBLISHED_INTERNAL` and `PUBLISHED_EXTERNAL`.   #### Draft When a Variable Definition is created it is assigned the status `DRAFT`. Under this status the Variable Definition is:  - Only visible to Statistics Norway employees. - Mutable (it may be changed directly without need for versioning). - Not suitable to refer to from other systems.  This status may be changed to `PUBLISHED_INTERNAL` or `PUBLISHED_EXTERNAL` with a direct update.  #### Published Internal Under this status the Variable Definition is:  - Only visible to Statistics Norway employees. - Immutable (all changes are versioned). - Suitable to refer to in internal systems for statistics production. - Not suitable to refer to for external use (for example in Statistikkbanken).  This status may be changed to `PUBLISHED_EXTERNAL` by creating a Patch version.  #### Published External Under this status the Variable Definition is:  - Visible to the general public. - Immutable (all changes are versioned). - Suitable to refer to from any system.  This status may not be changed as it would break immutability. If a Variable Definition is no longer relevant then its period of validity should be ended by specifying a `valid_until` date in a Patch version.  ### Immutability Variable Definitions are immutable. This means that any changes must be performed in a strict versioning system. Consumers can avoid being exposed to breaking changes by specifying a `date_of_validity` when they request a Variable Definition.  #### Patches Patches are for changes which do not affect the fundamental meaning of the Variable Definition.  #### Validity Periods Validity Periods are versions with a period defined by a `valid_from` date and optionally a `valid_until` date. If the fundamental meaning of a Variable Definition is to be changed, it should be done by creating a new Validity Period.

The version of the OpenAPI document: 0.1
Contact: metadata@ssb.no
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""

import io
import json
import re
import ssl

import urllib3

from .exceptions import ApiException
from .exceptions import ApiValueError

SUPPORTED_SOCKS_PROXIES = {"socks5", "socks5h", "socks4", "socks4a"}
RESTResponseType = urllib3.HTTPResponse


def is_socks_proxy_url(url):
    if url is None:
        return False
    split_section = url.split("://")
    if len(split_section) < 2:
        return False
    return split_section[0].lower() in SUPPORTED_SOCKS_PROXIES


class RESTResponse(io.IOBase):
    def __init__(self, resp) -> None:
        self.response = resp
        self.status = resp.status
        self.reason = resp.reason
        self.data = None

    def read(self):
        if self.data is None:
            self.data = self.response.data
        return self.data

    def getheaders(self):
        """Returns a dictionary of the response headers."""
        return self.response.headers

    def getheader(self, name, default=None):
        """Returns a given response header."""
        return self.response.headers.get(name, default)


class RESTClientObject:
    def __init__(self, configuration) -> None:
        # urllib3.PoolManager will pass all kw parameters to connectionpool
        # https://github.com/shazow/urllib3/blob/f9409436f83aeb79fbaf090181cd81b784f1b8ce/urllib3/poolmanager.py#L75
        # https://github.com/shazow/urllib3/blob/f9409436f83aeb79fbaf090181cd81b784f1b8ce/urllib3/connectionpool.py#L680
        # Custom SSL certificates and client certificates: http://urllib3.readthedocs.io/en/latest/advanced-usage.html

        # cert_reqs
        if configuration.verify_ssl:
            cert_reqs = ssl.CERT_REQUIRED
        else:
            cert_reqs = ssl.CERT_NONE

        pool_args = {
            "cert_reqs": cert_reqs,
            "ca_certs": configuration.ssl_ca_cert,
            "cert_file": configuration.cert_file,
            "key_file": configuration.key_file,
        }
        if configuration.assert_hostname is not None:
            pool_args["assert_hostname"] = configuration.assert_hostname

        if configuration.retries is not None:
            pool_args["retries"] = configuration.retries

        if configuration.tls_server_name:
            pool_args["server_hostname"] = configuration.tls_server_name

        if configuration.socket_options is not None:
            pool_args["socket_options"] = configuration.socket_options

        if configuration.connection_pool_maxsize is not None:
            pool_args["maxsize"] = configuration.connection_pool_maxsize

        # https pool manager
        self.pool_manager: urllib3.PoolManager

        if configuration.proxy:
            if is_socks_proxy_url(configuration.proxy):
                from urllib3.contrib.socks import SOCKSProxyManager

                pool_args["proxy_url"] = configuration.proxy
                pool_args["headers"] = configuration.proxy_headers
                self.pool_manager = SOCKSProxyManager(**pool_args)
            else:
                pool_args["proxy_url"] = configuration.proxy
                pool_args["proxy_headers"] = configuration.proxy_headers
                self.pool_manager = urllib3.ProxyManager(**pool_args)
        else:
            self.pool_manager = urllib3.PoolManager(**pool_args)

    def request(
        self,
        method,
        url,
        headers=None,
        body=None,
        post_params=None,
        _request_timeout=None,
    ):
        """Perform requests.

        :param method: http request method
        :param url: http request url
        :param headers: http request headers
        :param body: request json body, for `application/json`
        :param post_params: request post parameters,
                            `application/x-www-form-urlencoded`
                            and `multipart/form-data`
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        """
        method = method.upper()
        assert method in [
            "GET",
            "HEAD",
            "DELETE",
            "POST",
            "PUT",
            "PATCH",
            "OPTIONS",
        ]

        if post_params and body:
            raise ApiValueError(
                "body parameter cannot be used with post_params parameter.",
            )

        post_params = post_params or {}
        headers = headers or {}

        timeout = None
        if _request_timeout:
            if isinstance(_request_timeout, (int, float)):
                timeout = urllib3.Timeout(total=_request_timeout)
            elif isinstance(_request_timeout, tuple) and len(_request_timeout) == 2:
                timeout = urllib3.Timeout(
                    connect=_request_timeout[0],
                    read=_request_timeout[1],
                )

        try:
            # For `POST`, `PUT`, `PATCH`, `OPTIONS`, `DELETE`
            if method in ["POST", "PUT", "PATCH", "OPTIONS", "DELETE"]:
                # no content type provided or payload is json
                content_type = headers.get("Content-Type")
                if not content_type or re.search("json", content_type, re.IGNORECASE):
                    request_body = None
                    if body is not None:
                        request_body = json.dumps(body)
                    r = self.pool_manager.request(
                        method,
                        url,
                        body=request_body,
                        timeout=timeout,
                        headers=headers,
                        preload_content=False,
                    )
                elif content_type == "application/x-www-form-urlencoded":
                    r = self.pool_manager.request(
                        method,
                        url,
                        fields=post_params,
                        encode_multipart=False,
                        timeout=timeout,
                        headers=headers,
                        preload_content=False,
                    )
                elif content_type == "multipart/form-data":
                    # must del headers['Content-Type'], or the correct
                    # Content-Type which generated by urllib3 will be
                    # overwritten.
                    del headers["Content-Type"]
                    # Ensures that dict objects are serialized
                    post_params = [
                        (a, json.dumps(b)) if isinstance(b, dict) else (a, b)
                        for a, b in post_params
                    ]
                    r = self.pool_manager.request(
                        method,
                        url,
                        fields=post_params,
                        encode_multipart=True,
                        timeout=timeout,
                        headers=headers,
                        preload_content=False,
                    )
                # Pass a `string` parameter directly in the body to support
                # other content types than JSON when `body` argument is
                # provided in serialized form.
                elif isinstance(body, str) or isinstance(body, bytes):
                    r = self.pool_manager.request(
                        method,
                        url,
                        body=body,
                        timeout=timeout,
                        headers=headers,
                        preload_content=False,
                    )
                elif headers["Content-Type"].startswith("text/") and isinstance(
                    body, bool
                ):
                    request_body = "true" if body else "false"
                    r = self.pool_manager.request(
                        method,
                        url,
                        body=request_body,
                        preload_content=False,
                        timeout=timeout,
                        headers=headers,
                    )
                else:
                    # Cannot generate the request from given parameters
                    msg = """Cannot prepare a request message for provided
                             arguments. Please check that your arguments match
                             declared content type."""
                    raise ApiException(status=0, reason=msg)
            # For `GET`, `HEAD`
            else:
                r = self.pool_manager.request(
                    method,
                    url,
                    fields={},
                    timeout=timeout,
                    headers=headers,
                    preload_content=False,
                )
        except urllib3.exceptions.SSLError as e:
            msg = "\n".join([type(e).__name__, str(e)])
            raise ApiException(status=0, reason=msg)

        return RESTResponse(r)



"""Utilities for writing and reading existing variable definition files."""

import logging
from io import StringIO
from os import PathLike
from pathlib import Path
from typing import Any
from typing import TypeVar

from pydantic import BaseModel
from ruamel.yaml import YAML

from dapla_metadata.variable_definitions._generated.vardef_client.models.complete_response import (
    CompleteResponse,
)
from dapla_metadata.variable_definitions._utils.constants import HEADER
from dapla_metadata.variable_definitions._utils.files import _create_file_name
from dapla_metadata.variable_definitions._utils.files import _get_current_time
from dapla_metadata.variable_definitions._utils.files import (
    _model_to_yaml_with_comments,
)
from dapla_metadata.variable_definitions._utils.files import configure_yaml
from dapla_metadata.variable_definitions._utils.files import pre_process_data

logger = logging.getLogger(__name__)


T = TypeVar("T", bound=BaseModel)


def create_variable_yaml(
    model_instance: CompleteResponse,
    custom_directory: Path | None = None,
) -> Path:
    """Creates a yaml file for an existing variable definition."""
    file_name = _create_file_name(
        "variable_definition",
        _get_current_time(),
        model_instance.short_name,
        model_instance.id,
    )

    return _model_to_yaml_with_comments(
        model_instance,
        file_name,
        HEADER,
        custom_directory=custom_directory,
    )


def _read_variable_definition_file(file_path: Path) -> dict:
    yaml = YAML()
    configure_yaml(yaml)
    logger.debug("Full path to variable definition file %s", file_path)
    logger.info("Reading from '%s'", file_path.name)
    with file_path.open(encoding="utf-8") as f:
        return yaml.load(f)


def _strip_strings_recursively(data: Any) -> Any:
    """Recursively strip leading and trailing whitespace from string values in nested dicts/lists.

    This function traverses the provided data, which may be a dictionary, list, or other types,
    and applies the following logic:
        - If the data is a dictionary, it recursively strips string values in all key-value pairs.
        - If the data is a list, it recursively strips string values in all list elements.
        - If the data is a string, it strips leading and trailing whitespace.
        - Any other data types are returned unchanged.

    Args:
        data: The input data, which may include nested dictionaries, lists, or other types.

    Returns:
        Any: The processed data, with strings stripped of whitespace or unchanged if not a string.
    """
    if isinstance(data, dict):
        return {k: _strip_strings_recursively(v) for k, v in data.items()}
    if isinstance(data, list):
        return [_strip_strings_recursively(item) for item in data]
    if isinstance(data, str):
        return data.strip()
    return data


def _read_file_to_model(
    file_path: PathLike[str] | None,
    model_class: type[T],
) -> T:
    """Read from a variable definition file into the given Pydantic model.

    Args:
        file_path (PathLike[str]): The path to the file to read in.
        model_class (type[T]): The model to instantiate. Must inherit from Pydantic's BaseModel.

    Raises:
        TypeError: If no file path could be deduced.
        FileNotFoundError: If we could not instantiate the model.

    Returns:
        T: BaseModel: The instantiated Pydantic model
    """
    try:
        file_path = Path(
            # type incongruence (i.e. None) is handled by catching the exception
            file_path,  # type: ignore [arg-type]
        )
    except TypeError as e:
        msg = "Could not deduce a path to the file. Please supply a path to the yaml file you wish to submit with the `file_path` parameter."
        raise FileNotFoundError(
            msg,
        ) from e
    raw_data = _read_variable_definition_file(file_path)
    cleaned_data = _strip_strings_recursively(raw_data)

    model = model_class.from_dict(  # type:ignore [attr-defined]
        cleaned_data
    )

    if model is None:
        msg = f"Could not read data from {file_path}"
        raise FileNotFoundError(msg)
    return model


def _convert_to_yaml_output(model: BaseModel) -> str:
    """Convert a Pydantic model to YAML format.

    Args:
        model: A Pydantic model instance

    Returns:
        YAML string representation of the model
    """
    stream = StringIO()
    with YAML(output=stream) as yaml:
        configure_yaml(yaml)
        data = model.model_dump(
            mode="json",
            serialize_as_any=True,
            warnings="error",
        )
        yaml.dump(pre_process_data(data))
    return stream.getvalue()



from __future__ import annotations

import logging
from abc import ABC
from abc import abstractmethod
from typing import TYPE_CHECKING
from typing import Generic
from typing import TypeVar

if TYPE_CHECKING:
    from concurrent.futures import ThreadPoolExecutor

logger = logging.getLogger(__name__)

T = TypeVar("T")


class GetExternalSource(ABC, Generic[T]):
    """Abstract base class for retrieving data from external sources asynchronously.

    This class provides methods to initiate an asynchronous data retrieval
    operation, check its status, and retrieve the result once the operation
    completes. Subclasses must implement the `_fetch_data_from_external_source`
    method to define how data is fetched from the specific external source.
    """

    def __init__(self, executor: ThreadPoolExecutor) -> None:
        """Initialize the GetExternalSource with an executor to manage asynchronous tasks.

        This constructor initializes a future object that will hold the result of the
        asynchronous data fetching operation from an external source.

        Args:
            executor: An instance of ThreadPoolExecutor to manage the asynchronous
                execution of data fetching.
        """
        self.future = executor.submit(
            self._fetch_data_from_external_source,
        )

    def wait_for_external_result(self) -> None:
        """Wait for the thread responsible for loading the external request to finish.

        If there is no future to wait for, it logs a warning and returns immediately.
        """
        if not self.future:
            logger.warning("No future to wait for.")
            return
        self.future.result()

    def check_if_external_data_is_loaded(self) -> bool:
        """Check if the thread getting the external data has finished running.

        Returns:
            True if the data fetching operation is complete, False otherwise.
        """
        if self.future:
            return self.future.done()
        return False

    def retrieve_external_data(self) -> T | None:
        """Retrieve the result of the data fetching operation.

        This method checks if the asynchronous data fetching operation has
        completed. If the operation is finished, it returns the result.
        Otherwise, it returns None.

        Returns:
            The result of the data fetching operation if it is complete or None
            if the operation has not yet finished.
        """
        if self.future:
            return self.future.result()
        return None

    @abstractmethod
    def _fetch_data_from_external_source(self) -> T | None:
        """Handle external data retrieval.

        Abstract method to be implemented in the subclass.
        This method should define the logic for retrieving data from the specific
        external source.

        Returns:
            The data retrieved from the external source.
        """
        raise NotImplementedError



"""Variable Definitions

## Introduction  Variable Definitions are centralized definitions of concrete variables which are typically present in multiple datasets. Variable Definitions support standardization of data and metadata and facilitate sharing and joining of data by clarifying when variables have an identical definition.  ## Maintenance of Variable Definitions This API allows for creation, maintenance and access of Variable Definitions.  ### Ownership Creation and maintenance of variables may only be performed by Statistics Norway employees representing a specific Dapla team, who are defined as the owners of a given Variable Definition. The team an owner represents must be specified when making a request through the `active_group` query parameter. All maintenance is to be performed by the owners, with no intervention from administrators.  ### Status All Variable Definitions have an associated status. The possible values for status are `DRAFT`, `PUBLISHED_INTERNAL` and `PUBLISHED_EXTERNAL`.   #### Draft When a Variable Definition is created it is assigned the status `DRAFT`. Under this status the Variable Definition is:  - Only visible to Statistics Norway employees. - Mutable (it may be changed directly without need for versioning). - Not suitable to refer to from other systems.  This status may be changed to `PUBLISHED_INTERNAL` or `PUBLISHED_EXTERNAL` with a direct update.  #### Published Internal Under this status the Variable Definition is:  - Only visible to Statistics Norway employees. - Immutable (all changes are versioned). - Suitable to refer to in internal systems for statistics production. - Not suitable to refer to for external use (for example in Statistikkbanken).  This status may be changed to `PUBLISHED_EXTERNAL` by creating a Patch version.  #### Published External Under this status the Variable Definition is:  - Visible to the general public. - Immutable (all changes are versioned). - Suitable to refer to from any system.  This status may not be changed as it would break immutability. If a Variable Definition is no longer relevant then its period of validity should be ended by specifying a `valid_until` date in a Patch version.  ### Immutability Variable Definitions are immutable. This means that any changes must be performed in a strict versioning system. Consumers can avoid being exposed to breaking changes by specifying a `date_of_validity` when they request a Variable Definition.  #### Patches Patches are for changes which do not affect the fundamental meaning of the Variable Definition.  #### Validity Periods Validity Periods are versions with a period defined by a `valid_from` date and optionally a `valid_until` date. If the fundamental meaning of a Variable Definition is to be changed, it should be done by creating a new Validity Period.

The version of the OpenAPI document: 0.1
Contact: metadata@ssb.no
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""

from typing import Any

from typing_extensions import Self


class OpenApiException(Exception):
    """The base exception class for all OpenAPIExceptions"""


class ApiTypeError(OpenApiException, TypeError):
    def __init__(
        self, msg, path_to_item=None, valid_classes=None, key_type=None
    ) -> None:
        """Raises an exception for TypeErrors

        Args:
            msg (str): the exception message

        Keyword Args:
            path_to_item (list): a list of keys an indices to get to the
                                 current_item
                                 None if unset
            valid_classes (tuple): the primitive classes that current item
                                   should be an instance of
                                   None if unset
            key_type (bool): False if our value is a value in a dict
                             True if it is a key in a dict
                             False if our item is an item in a list
                             None if unset
        """
        self.path_to_item = path_to_item
        self.valid_classes = valid_classes
        self.key_type = key_type
        full_msg = msg
        if path_to_item:
            full_msg = f"{msg} at {render_path(path_to_item)}"
        super(ApiTypeError, self).__init__(full_msg)


class ApiValueError(OpenApiException, ValueError):
    def __init__(self, msg, path_to_item=None) -> None:
        """Args:
            msg (str): the exception message

        Keyword Args:
            path_to_item (list) the path to the exception in the
                received_data dict. None if unset
        """
        self.path_to_item = path_to_item
        full_msg = msg
        if path_to_item:
            full_msg = f"{msg} at {render_path(path_to_item)}"
        super(ApiValueError, self).__init__(full_msg)


class ApiAttributeError(OpenApiException, AttributeError):
    def __init__(self, msg, path_to_item=None) -> None:
        """Raised when an attribute reference or assignment fails.

        Args:
            msg (str): the exception message

        Keyword Args:
            path_to_item (None/list) the path to the exception in the
                received_data dict
        """
        self.path_to_item = path_to_item
        full_msg = msg
        if path_to_item:
            full_msg = f"{msg} at {render_path(path_to_item)}"
        super(ApiAttributeError, self).__init__(full_msg)


class ApiKeyError(OpenApiException, KeyError):
    def __init__(self, msg, path_to_item=None) -> None:
        """Args:
            msg (str): the exception message

        Keyword Args:
            path_to_item (None/list) the path to the exception in the
                received_data dict
        """
        self.path_to_item = path_to_item
        full_msg = msg
        if path_to_item:
            full_msg = f"{msg} at {render_path(path_to_item)}"
        super(ApiKeyError, self).__init__(full_msg)


class ApiException(OpenApiException):
    def __init__(
        self,
        status=None,
        reason=None,
        http_resp=None,
        *,
        body: str | None = None,
        data: Any | None = None,
    ) -> None:
        self.status = status
        self.reason = reason
        self.body = body
        self.data = data
        self.headers = None

        if http_resp:
            if self.status is None:
                self.status = http_resp.status
            if self.reason is None:
                self.reason = http_resp.reason
            if self.body is None:
                try:
                    self.body = http_resp.data.decode("utf-8")
                except Exception:
                    pass
            self.headers = http_resp.getheaders()

    @classmethod
    def from_response(
        cls,
        *,
        http_resp,
        body: str | None,
        data: Any | None,
    ) -> Self:
        if http_resp.status == 400:
            raise BadRequestException(http_resp=http_resp, body=body, data=data)

        if http_resp.status == 401:
            raise UnauthorizedException(http_resp=http_resp, body=body, data=data)

        if http_resp.status == 403:
            raise ForbiddenException(http_resp=http_resp, body=body, data=data)

        if http_resp.status == 404:
            raise NotFoundException(http_resp=http_resp, body=body, data=data)

        if 500 <= http_resp.status <= 599:
            raise ServiceException(http_resp=http_resp, body=body, data=data)
        raise ApiException(http_resp=http_resp, body=body, data=data)

    def __str__(self):
        """Custom error messages for exception"""
        error_message = f"({self.status})\nReason: {self.reason}\n"
        if self.headers:
            error_message += f"HTTP response headers: {self.headers}\n"

        if self.data or self.body:
            error_message += f"HTTP response body: {self.data or self.body}\n"

        return error_message


class BadRequestException(ApiException):
    pass


class NotFoundException(ApiException):
    pass


class UnauthorizedException(ApiException):
    pass


class ForbiddenException(ApiException):
    pass


class ServiceException(ApiException):
    pass


def render_path(path_to_item):
    """Returns a string representation of a path"""
    result = ""
    for pth in path_to_item:
        if isinstance(pth, int):
            result += f"[{pth}]"
        else:
            result += f"['{pth}']"
    return result



"""Variable Definitions

## Introduction  Variable Definitions are centralized definitions of concrete variables which are typically present in multiple datasets. Variable Definitions support standardization of data and metadata and facilitate sharing and joining of data by clarifying when variables have an identical definition.  ## Maintenance of Variable Definitions This API allows for creation, maintenance and access of Variable Definitions.  ### Ownership Creation and maintenance of variables may only be performed by Statistics Norway employees representing a specific Dapla team, who are defined as the owners of a given Variable Definition. The team an owner represents must be specified when making a request through the `active_group` query parameter. All maintenance is to be performed by the owners, with no intervention from administrators.  ### Status All Variable Definitions have an associated status. The possible values for status are `DRAFT`, `PUBLISHED_INTERNAL` and `PUBLISHED_EXTERNAL`.   #### Draft When a Variable Definition is created it is assigned the status `DRAFT`. Under this status the Variable Definition is:  - Only visible to Statistics Norway employees. - Mutable (it may be changed directly without need for versioning). - Not suitable to refer to from other systems.  This status may be changed to `PUBLISHED_INTERNAL` or `PUBLISHED_EXTERNAL` with a direct update.  #### Published Internal Under this status the Variable Definition is:  - Only visible to Statistics Norway employees. - Immutable (all changes are versioned). - Suitable to refer to in internal systems for statistics production. - Not suitable to refer to for external use (for example in Statistikkbanken).  This status may be changed to `PUBLISHED_EXTERNAL` by creating a Patch version.  #### Published External Under this status the Variable Definition is:  - Visible to the general public. - Immutable (all changes are versioned). - Suitable to refer to from any system.  This status may not be changed as it would break immutability. If a Variable Definition is no longer relevant then its period of validity should be ended by specifying a `valid_until` date in a Patch version.  ### Immutability Variable Definitions are immutable. This means that any changes must be performed in a strict versioning system. Consumers can avoid being exposed to breaking changes by specifying a `date_of_validity` when they request a Variable Definition.  #### Patches Patches are for changes which do not affect the fundamental meaning of the Variable Definition.  #### Validity Periods Validity Periods are versions with a period defined by a `valid_from` date and optionally a `valid_until` date. If the fundamental meaning of a Variable Definition is to be changed, it should be done by creating a new Validity Period.

The version of the OpenAPI document: 0.1
Contact: metadata@ssb.no
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""

import copy
import http.client as httplib
import logging
import multiprocessing
import sys
from logging import FileHandler

import urllib3

JSON_SCHEMA_VALIDATION_KEYWORDS = {
    "multipleOf",
    "maximum",
    "exclusiveMaximum",
    "minimum",
    "exclusiveMinimum",
    "maxLength",
    "minLength",
    "pattern",
    "maxItems",
    "minItems",
}


class Configuration:
    """This class contains various settings of the API client.

    :param host: Base url.
    :param ignore_operation_servers
      Boolean to ignore operation servers for the API client.
      Config will use `host` as the base url regardless of the operation servers.
    :param api_key: Dict to store API key(s).
      Each entry in the dict specifies an API key.
      The dict key is the name of the security scheme in the OAS specification.
      The dict value is the API key secret.
    :param api_key_prefix: Dict to store API prefix (e.g. Bearer).
      The dict key is the name of the security scheme in the OAS specification.
      The dict value is an API key prefix when generating the auth data.
    :param username: Username for HTTP basic authentication.
    :param password: Password for HTTP basic authentication.
    :param access_token: Access token.
    :param server_index: Index to servers configuration.
    :param server_variables: Mapping with string values to replace variables in
      templated server configuration. The validation of enums is performed for
      variables with defined enum values before.
    :param server_operation_index: Mapping from operation ID to an index to server
      configuration.
    :param server_operation_variables: Mapping from operation ID to a mapping with
      string values to replace variables in templated server configuration.
      The validation of enums is performed for variables with defined enum
      values before.
    :param ssl_ca_cert: str - the path to a file of concatenated CA certificates
      in PEM format.
    :param retries: Number of retries for API requests.

    :Example:
    """

    _default = None

    def __init__(
        self,
        host=None,
        api_key=None,
        api_key_prefix=None,
        username=None,
        password=None,
        access_token=None,
        server_index=None,
        server_variables=None,
        server_operation_index=None,
        server_operation_variables=None,
        ignore_operation_servers=False,
        ssl_ca_cert=None,
        retries=None,
        *,
        debug: bool | None = None,
    ) -> None:
        """Constructor"""
        self._base_path = (
            "https://metadata.intern.test.ssb.no" if host is None else host
        )
        """Default Base url
        """
        self.server_index = 0 if server_index is None and host is None else server_index
        self.server_operation_index = server_operation_index or {}
        """Default server index
        """
        self.server_variables = server_variables or {}
        self.server_operation_variables = server_operation_variables or {}
        """Default server variables
        """
        self.ignore_operation_servers = ignore_operation_servers
        """Ignore operation servers
        """
        self.temp_folder_path = None
        """Temp file folder for downloading files
        """
        # Authentication Settings
        self.api_key = {}
        if api_key:
            self.api_key = api_key
        """dict to store API key(s)
        """
        self.api_key_prefix = {}
        if api_key_prefix:
            self.api_key_prefix = api_key_prefix
        """dict to store API prefix (e.g. Bearer)
        """
        self.refresh_api_key_hook = None
        """function hook to refresh API key if expired
        """
        self.username = username
        """Username for HTTP basic authentication
        """
        self.password = password
        """Password for HTTP basic authentication
        """
        self.access_token = access_token
        """Access token
        """
        self.logger = {}
        """Logging Settings
        """
        self.logger["package_logger"] = logging.getLogger("vardef_client")
        self.logger["urllib3_logger"] = logging.getLogger("urllib3")
        self.logger_format = "%(asctime)s %(levelname)s %(message)s"
        """Log format
        """
        self.logger_stream_handler = None
        """Log stream handler
        """
        self.logger_file_handler: FileHandler | None = None
        """Log file handler
        """
        self.logger_file = None
        """Debug file location
        """
        if debug is not None:
            self.debug = debug
        else:
            self.__debug = False
        """Debug switch
        """

        self.verify_ssl = True
        """SSL/TLS verification
           Set this to false to skip verifying SSL certificate when calling API
           from https server.
        """
        self.ssl_ca_cert = ssl_ca_cert
        """Set this to customize the certificate file to verify the peer.
        """
        self.cert_file = None
        """client certificate file
        """
        self.key_file = None
        """client key file
        """
        self.assert_hostname = None
        """Set this to True/False to enable/disable SSL hostname verification.
        """
        self.tls_server_name = None
        """SSL/TLS Server Name Indication (SNI)
           Set this to the SNI value expected by the server.
        """

        self.connection_pool_maxsize = multiprocessing.cpu_count() * 5
        """urllib3 connection pool's maximum number of connections saved
           per pool. urllib3 uses 1 connection as default value, but this is
           not the best value when you are making a lot of possibly parallel
           requests to the same host, which is often the case here.
           cpu_count * 5 is used as default value to increase performance.
        """

        self.proxy: str | None = None
        """Proxy URL
        """
        self.proxy_headers = None
        """Proxy headers
        """
        self.safe_chars_for_path_param = ""
        """Safe chars for path_param
        """
        self.retries = retries
        """Adding retries to override urllib3 default value 3
        """
        # Enable client side validation
        self.client_side_validation = True

        self.socket_options = None
        """Options to pass down to the underlying urllib3 socket
        """

        self.datetime_format = "%Y-%m-%dT%H:%M:%S.%f%z"
        """datetime format
        """

        self.date_format = "%Y-%m-%d"
        """date format
        """

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if k not in ("logger", "logger_file_handler"):
                setattr(result, k, copy.deepcopy(v, memo))
        # shallow copy of loggers
        result.logger = copy.copy(self.logger)
        # use setters to configure loggers
        result.logger_file = self.logger_file
        result.debug = self.debug
        return result

    def __setattr__(self, name, value):
        object.__setattr__(self, name, value)

    @classmethod
    def set_default(cls, default):
        """Set default instance of configuration.

        It stores default configuration, which can be
        returned by get_default_copy method.

        :param default: object of Configuration
        """
        cls._default = default

    @classmethod
    def get_default_copy(cls):
        """Deprecated. Please use `get_default` instead.

        Deprecated. Please use `get_default` instead.

        :return: The configuration object.
        """
        return cls.get_default()

    @classmethod
    def get_default(cls):
        """Return the default configuration.

        This method returns newly created, based on default constructor,
        object of Configuration class or returns a copy of default
        configuration.

        :return: The configuration object.
        """
        if cls._default is None:
            cls._default = Configuration()
        return cls._default

    @property
    def logger_file(self):
        """The logger file.

        If the logger_file is None, then add stream handler and remove file
        handler. Otherwise, add file handler and remove stream handler.

        :param value: The logger_file path.
        :type: str
        """
        return self.__logger_file

    @logger_file.setter
    def logger_file(self, value):
        """The logger file.

        If the logger_file is None, then add stream handler and remove file
        handler. Otherwise, add file handler and remove stream handler.

        :param value: The logger_file path.
        :type: str
        """
        self.__logger_file = value
        if self.__logger_file:
            # If set logging file,
            # then add file handler and remove stream handler.
            self.logger_file_handler = logging.FileHandler(self.__logger_file)
            self.logger_file_handler.setFormatter(self.logger_formatter)
            for _, logger in self.logger.items():
                logger.addHandler(self.logger_file_handler)

    @property
    def debug(self):
        """Debug status

        :param value: The debug status, True or False.
        :type: bool
        """
        return self.__debug

    @debug.setter
    def debug(self, value):
        """Debug status

        :param value: The debug status, True or False.
        :type: bool
        """
        self.__debug = value
        if self.__debug:
            # if debug status is True, turn on debug logging
            for _, logger in self.logger.items():
                logger.setLevel(logging.DEBUG)
            # turn on httplib debug
            httplib.HTTPConnection.debuglevel = 1
        else:
            # if debug status is False, turn off debug logging,
            # setting log level to default `logging.WARNING`
            for _, logger in self.logger.items():
                logger.setLevel(logging.WARNING)
            # turn off httplib debug
            httplib.HTTPConnection.debuglevel = 0

    @property
    def logger_format(self):
        """The logger format.

        The logger_formatter will be updated when sets logger_format.

        :param value: The format string.
        :type: str
        """
        return self.__logger_format

    @logger_format.setter
    def logger_format(self, value):
        """The logger format.

        The logger_formatter will be updated when sets logger_format.

        :param value: The format string.
        :type: str
        """
        self.__logger_format = value
        self.logger_formatter = logging.Formatter(self.__logger_format)

    def get_api_key_with_prefix(self, identifier, alias=None):
        """Gets API key (with prefix if set).

        :param identifier: The identifier of apiKey.
        :param alias: The alternative identifier of apiKey.
        :return: The token for api key authentication.
        """
        if self.refresh_api_key_hook is not None:
            self.refresh_api_key_hook(self)
        key = self.api_key.get(
            identifier, self.api_key.get(alias) if alias is not None else None
        )
        if key:
            prefix = self.api_key_prefix.get(identifier)
            if prefix:
                return "%s %s" % (prefix, key)
            return key

    def get_basic_auth_token(self):
        """Gets HTTP basic authentication header (string).

        :return: The token for basic HTTP authentication.
        """
        username = ""
        if self.username is not None:
            username = self.username
        password = ""
        if self.password is not None:
            password = self.password
        return urllib3.util.make_headers(
            basic_auth=username + ":" + password,
        ).get("authorization")

    def auth_settings(self):
        """Gets Auth Settings dict for api client.

        :return: The Auth Settings information dict.
        """
        auth = {}
        if self.access_token is not None:
            auth["keycloak_token"] = {
                "type": "bearer",
                "in": "header",
                "format": "JWT",
                "key": "Authorization",
                "value": "Bearer " + self.access_token,
            }
        return auth

    def to_debug_report(self):
        """Gets the essential information for debugging.

        :return: The report for debugging.
        """
        return (
            "Python SDK Debug Report:\n"
            f"OS: {sys.platform}\n"
            f"Python Version: {sys.version}\n"
            "Version of the API: 0.1\n"
            "SDK Package Version: 1.0.0"
        )

    def get_host_settings(self):
        """Gets an array of host settings

        :return: An array of host settings
        """
        return [
            {
                "url": "https://metadata.intern.test.ssb.no",
                "description": "Internal test server",
            },
        ]

    def get_host_from_settings(self, index, variables=None, servers=None):
        """Gets host URL based on the index and variables
        :param index: array index of the host settings
        :param variables: hash of variable and the corresponding value
        :param servers: an array of host settings or None
        :return: URL based on host settings
        """
        if index is None:
            return self._base_path

        variables = {} if variables is None else variables
        servers = self.get_host_settings() if servers is None else servers

        try:
            server = servers[index]
        except IndexError:
            raise ValueError(
                f"Invalid index {index} when selecting the host settings. "
                f"Must be less than {len(servers)}"
            )

        url = server["url"]

        # go through variables and replace placeholders
        for variable_name, variable in server.get("variables", {}).items():
            used_value = variables.get(variable_name, variable["default_value"])

            if "enum_values" in variable and used_value not in variable["enum_values"]:
                raise ValueError(
                    "The variable `{0}` in the host URL has invalid value "
                    "{1}. Must be {2}.".format(
                        variable_name, variables[variable_name], variable["enum_values"]
                    )
                )

            url = url.replace("{" + variable_name + "}", used_value)

        return url

    @property
    def host(self):
        """Return generated host."""
        return self.get_host_from_settings(
            self.server_index, variables=self.server_variables
        )

    @host.setter
    def host(self, value):
        """Fix base path."""
        self._base_path = value
        self.server_index = None



from pathlib import Path
from typing import TYPE_CHECKING

from dapla_metadata.variable_definitions._generated.vardef_client.models.complete_response import (
    CompleteResponse,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.contact import (
    Contact,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.language_string_type import (
    LanguageStringType,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.owner import (
    Owner,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.variable_status import (
    VariableStatus,
)
from dapla_metadata.variable_definitions._utils.constants import DEFAULT_DATE
from dapla_metadata.variable_definitions._utils.constants import TEMPLATE_HEADER
from dapla_metadata.variable_definitions._utils.files import _create_file_name
from dapla_metadata.variable_definitions._utils.files import _get_current_time
from dapla_metadata.variable_definitions._utils.files import (
    _get_variable_definitions_dir,
)
from dapla_metadata.variable_definitions._utils.files import (
    _model_to_yaml_with_comments,
)
from dapla_metadata.variable_definitions._utils.files import logger

if TYPE_CHECKING:
    from dapla_metadata.variable_definitions.variable_definition import (
        VariableDefinition,
    )


def _get_default_template() -> "VariableDefinition":
    from dapla_metadata.variable_definitions.variable_definition import (
        VariableDefinition,
    )

    return VariableDefinition(
        name=LanguageStringType(
            nb="Navn",
        ),
        short_name="generert_kortnavn",
        definition=LanguageStringType(
            nb="Definisjonstekst",
        ),
        valid_from=DEFAULT_DATE,
        unit_types=[""],
        subject_fields=[""],
        contains_special_categories_of_personal_data=False,
        owner=Owner(team="default team", groups=["default group"]),
        contact=Contact(
            title=LanguageStringType(
                nb="generert tittel",
            ),
            email="generert@ssb.no",
        ),
        variable_status=VariableStatus.DRAFT.value,
        id="",
        patch_id=0,
        created_at=DEFAULT_DATE,
        created_by="",
        last_updated_at=DEFAULT_DATE,
        last_updated_by="",
    )


def create_template_yaml(
    model_instance: CompleteResponse | None = None,
    custom_directory: Path | None = None,
) -> Path:
    """Creates a template yaml file for a new variable definition."""
    if model_instance is None:
        model_instance = _get_default_template()
    file_name = _create_file_name(
        "variable_definition_template",
        _get_current_time(),
    )

    file_path = _model_to_yaml_with_comments(
        model_instance,
        file_name,
        TEMPLATE_HEADER,
        custom_directory=custom_directory,
    )
    logger.debug("Created %s", file_path)
    return file_path


def _find_latest_template_file(directory: Path | None = None) -> Path | None:
    def _filter_template_file(path: Path) -> bool:
        return "variable_definition_template" in path.stem and path.suffix == ".yaml"

    try:
        return sorted(
            filter(
                _filter_template_file,
                (directory or _get_variable_definitions_dir()).iterdir(),
            ),
        )[-1]
    except IndexError:
        return None



import logging
from pathlib import Path

from pydantic import ConfigDict
from pydantic import PrivateAttr

from dapla_metadata.variable_definitions._generated.vardef_client.models.vardok_vardef_id_pair_response import (
    VardokVardefIdPairResponse,
)
from dapla_metadata.variable_definitions._utils.variable_definition_files import (
    _convert_to_yaml_output,
)

logger = logging.getLogger(__name__)


class VardokVardefIdPair(VardokVardefIdPairResponse):
    """A Vardok id.

    - Provides access to the fields of a Vardok Vardef id pair.
    - Provides methods allowing for nicer output of the Vardok Vardef id pair.

    Args:
        VardokVardefIdPairResponse: The Pydantic model superclass, representing a Vardok Vardef id pair response.
    """

    _file_path: Path | None = PrivateAttr(None)
    model_config = ConfigDict(use_enum_values=True, str_strip_whitespace=True)

    @staticmethod
    def from_model(
        model: VardokVardefIdPairResponse,
    ) -> "VardokVardefIdPair":
        """Create a VardokVardefIdPair instance from a VardokVardefIdPairResponse."""
        return VardokVardefIdPair.model_construct(**model.model_dump())

    def __str__(self) -> str:
        """Format as indented YAML."""
        return _convert_to_yaml_output(self)

    def __repr__(self) -> str:
        """Format as indented YAML."""
        return _convert_to_yaml_output(self)



"""Enumerations used in Datadoc."""

from __future__ import annotations

from enum import Enum


class SupportedLanguages(str, Enum):
    """The list of languages metadata may be recorded in.

    Reference: https://www.iana.org/assignments/language-subtag-registry/language-subtag-registry
    """

    NORSK_BOKMÅL = "nb"  # noqa: PLC2401 the listed problems do not apply in this case
    NORSK_NYNORSK = "nn"
    ENGLISH = "en"



"""Upgrade old metadata files to be compatible with new versions.

An important principle of Datadoc is that we ALWAYS guarantee backwards
compatibility of existing metadata documents. This means that we guarantee
that a user will never lose data, even if their document is decades old.

For each document version we release with breaking changes, we implement a
handler and register the version by defining a BackwardsCompatibleVersion
instance. These documents will then be upgraded when they're opened in Datadoc.

A test must also be implemented for each new version.
"""

from __future__ import annotations

import logging
from collections import OrderedDict
from dataclasses import dataclass
from datetime import datetime
from datetime import timezone
from typing import TYPE_CHECKING
from typing import Any

import arrow

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from collections.abc import Callable

VERSION_FIELD_NAME = "document_version"
PSEUDONYMIZATION_KEY = "pseudonymization"


class UnknownModelVersionError(Exception):
    """Exception raised for unknown model versions.

    This error is thrown when an unrecognized model version is encountered.
    """

    def __init__(
        self,
        supplied_version: str,
        *args: tuple[Any, ...],
    ) -> None:
        """Initialize the exception with the supplied version.

        Args:
            supplied_version: The version of the model that was not recognized.
            *args: Additional arguments for the Exception base class.
        """
        super().__init__(args)
        self.supplied_version = supplied_version

    def __str__(self) -> str:
        """Return string representation."""
        return f"Document Version ({self.supplied_version}) of discovered file is not supported"


SUPPORTED_VERSIONS: OrderedDict[str, BackwardsCompatibleVersion] = OrderedDict()


@dataclass()
class BackwardsCompatibleVersion:
    """A version which we support with backwards compatibility.

    This class registers a version and its corresponding handler function
    for backwards compatibility.
    """

    version: str
    handler: Callable[[dict[str, Any]], dict[str, Any]]

    def __post_init__(self) -> None:
        """Register this version in the supported versions map.

        This method adds the instance to the `SUPPORTED_VERSIONS` dictionary
        using the version as the key.
        """
        SUPPORTED_VERSIONS[self.version] = self


def handle_current_version(supplied_metadata: dict[str, Any]) -> dict[str, Any]:
    """Handle the current version of the metadata.

    This function returns the supplied metadata unmodified.

    Args:
        supplied_metadata: The metadata for the current version.

    Returns:
        The unmodified supplied metadata.
    """
    return supplied_metadata


def _find_and_update_language_strings(supplied_metadata: dict | None) -> dict | None:
    """Find and update language-specific strings in the supplied metadata.

    This function iterates through the supplied metadata dictionary.
    For each key-value pair, if the value is a dictionary containing "en"
    it is passed to the `_convert_language_string_type` function to potentially
    update its format.

    Args:
        supplied_metadata: A metadata dictionary where values may include nested
            dictionaries with language-specific strings.

    Returns:
        The updated metadata dictionary. If the supplied metadata is not a
        dictionary, it returns `None`.
    """
    if isinstance(supplied_metadata, dict):
        for key, value in supplied_metadata.items():
            if isinstance(value, dict) and "en" in value:
                supplied_metadata[key] = _convert_language_string_type(value)
        return supplied_metadata
    return None


def _convert_language_string_type(supplied_value: dict) -> list[dict[str, str]]:
    """Convert a dictionary of language-specific strings to a list of dictionaries.

    This function takes a dictionary with language codes as keys and
    corresponding language-specific strings as values, and converts it to a list
    of dictionaries with 'languageCode' and 'languageText' keys.

    Args:
        supplied_value: A dictionary containing language codes as keys and
            language strings as values.

    Returns:
        A list of dictionaries, each containing 'languageCode' and 'languageText'
        keys, representing the converted language strings.
    """
    return [
        {
            "languageCode": "en",
            "languageText": supplied_value["en"],
        },
        {
            "languageCode": "nn",
            "languageText": supplied_value["nn"],
        },
        {
            "languageCode": "nb",
            "languageText": supplied_value["nb"],
        },
    ]


def _remove_element_from_model(
    supplied_metadata: dict[str, Any],
    element_to_remove: str,
) -> None:
    """Remove an element from the supplied metadata dictionary.

    This function deletes a specified element from the supplied metadata dictionary
    if it exists.

    Args:
        supplied_metadata: The metadata dictionary from which the element will be
            removed.
        element_to_remove: The key of the element to be removed from the metadata
            dictionary.
    """
    supplied_metadata.pop(element_to_remove, None)


def _cast_to_date_type(value_to_update: str | None) -> str | None:
    """Convert a string to a date string in ISO format.

    This function takes a string representing a date and converts it to a
    date string in ISO format. If the input is `None`, it returns `None` without
    modification.

    Args:
        value_to_update: A string representing a date or `None`.

    Returns:
        The date string in ISO format if the input was a valid date string, or
        `None` if the input was `None`.
    """
    if value_to_update is None:
        return value_to_update

    return str(
        arrow.get(
            value_to_update,
        ).date(),
    )


def convert_is_personal_data(supplied_metadata: dict[str, Any]) -> None:
    """Convert 'is_personal_data' values in the supplied metadata to boolean.

    Iterates over variables in the supplied metadata and updates the
    'is_personal_data' field:
      - Sets it to True for NON_PSEUDONYMISED_ENCRYPTED_PERSONAL_DATA and PSEUDONYMISED_ENCRYPTED_PERSONAL_DATA.
      - Sets it to False for NOT_PERSONAL_DATA.

    Args:
        supplied_metadata: The metadata dictionary to be updated.
    """
    for variable in supplied_metadata["datadoc"]["variables"]:
        value = variable["is_personal_data"]
        if value in (
            "NON_PSEUDONYMISED_ENCRYPTED_PERSONAL_DATA",
            "PSEUDONYMISED_ENCRYPTED_PERSONAL_DATA",
        ):
            variable["is_personal_data"] = True
        elif value == "NOT_PERSONAL_DATA":
            variable["is_personal_data"] = False


def copy_pseudonymization_metadata(supplied_metadata: dict[str, Any]) -> None:
    """Copies pseudonymization metadata from the old pseudonymization section into the corresponding variable.

    For each variable in `supplied_metadata["datadoc"]["variables"]` that has a matching
    `short_name` in `supplied_metadata["pseudonymization"]["pseudo_variables"]`, this
    function copies the following fields into the variable's 'pseudonymization' dictionary:

        - stable_identifier_type
        - stable_identifier_version
        - encryption_algorithm
        - encryption_key_reference
        - encryption_algorithm_parameters

    From the pseudo_dataset the value dataset_pseudo_time is copied to each variable as pseudonymization_time.

    Args:
        supplied_metadata: The metadata dictionary to be updated.
    """
    pseudo_vars = supplied_metadata.get(PSEUDONYMIZATION_KEY, {}).get(
        "pseudo_variables", []
    )
    pseudo_dataset = (
        supplied_metadata.get(PSEUDONYMIZATION_KEY, {}).get("pseudo_dataset") or {}
    )
    pseudo_time = pseudo_dataset.get("dataset_pseudo_time", None)
    datadoc_vars = supplied_metadata.get("datadoc", {}).get("variables", [])
    pseudo_lookup = {var.get("short_name"): var for var in pseudo_vars}

    for variable in datadoc_vars:
        short_name = variable.get("short_name")
        if short_name in pseudo_lookup:
            pseudo_var = pseudo_lookup[short_name]
            variable[PSEUDONYMIZATION_KEY] = variable.get(
                PSEUDONYMIZATION_KEY, {}
            ).copy()

            for field in [
                "stable_identifier_type",
                "stable_identifier_version",
                "encryption_algorithm",
                "encryption_key_reference",
                "encryption_algorithm_parameters",
            ]:
                variable[PSEUDONYMIZATION_KEY][field] = pseudo_var[field]
            variable[PSEUDONYMIZATION_KEY]["pseudonymization_time"] = pseudo_time

        else:
            variable[PSEUDONYMIZATION_KEY] = None


def handle_version_4_0_0(supplied_metadata: dict[str, Any]) -> dict[str, Any]:
    """Handle breaking changes for version 5.0.1.

    This function modifies the supplied metadata to accommodate breaking changes
    introduced in version 5.0.1. Specifically, it:
    - Copies pseudonymization metadata if pseudonymization is enabled.
    - Converts the 'is_personal_data' fields to be a bool.
    - Updates the 'document_version' field in the 'datadoc' section to "5.0.1".
    - All 'pseudonymization' from the container is removed.
    - It also updates the container version to 1.0.0 from 0.0.1

    Args:
        supplied_metadata: The metadata dictionary to be updated.

    Returns:
        The updated metadata dictionary.
    """
    if supplied_metadata.get(PSEUDONYMIZATION_KEY):
        copy_pseudonymization_metadata(supplied_metadata)

    convert_is_personal_data(supplied_metadata)

    supplied_metadata["datadoc"]["document_version"] = "5.0.1"
    _remove_element_from_model(supplied_metadata, PSEUDONYMIZATION_KEY)
    supplied_metadata["document_version"] = "1.0.0"
    return supplied_metadata


def handle_version_3_3_0(supplied_metadata: dict[str, Any]) -> dict[str, Any]:
    """Handle breaking changes for version 3.3.0.

    This function modifies the supplied metadata to accommodate breaking changes
    introduced in version 4.0.0. Specifically, it removes the
    'direct_person_identifying' field from each variable in 'datadoc.variables'
    and updates the 'document_version' field to "4.0.0".

    Version 4.0.0 used an enum for is_personal_data, however this was changed to a bool again for version 5.0.1.
    We skip setting the enum here and just keep the value it has.

    Args:
        supplied_metadata: The metadata dictionary to be updated.

    Returns:
        The updated metadata dictionary.
    """
    for variable in supplied_metadata["datadoc"]["variables"]:
        variable["is_personal_data"] = variable["direct_person_identifying"]
        _remove_element_from_model(variable, "direct_person_identifying")

    supplied_metadata["datadoc"]["document_version"] = "4.0.0"
    return supplied_metadata


def handle_version_3_2_0(supplied_metadata: dict[str, Any]) -> dict[str, Any]:
    """Handle breaking changes for version 3.2.0.

    This function modifies the supplied metadata to accommodate breaking
    changes introduced in version 3.3.0. Specifically, it updates the
    'contains_data_from' and 'contains_data_until' fields in both the 'dataset'
    and 'variables' sections of the supplied metadata dictionary to ensure they
    are stored as date strings.
    It also updates the 'document_version' field to "3.3.0".

    Args:
        supplied_metadata: The metadata dictionary to be updated.

    Returns:
        The updated metadata dictionary.
    """
    fields = ["contains_data_from", "contains_data_until"]
    for field in fields:
        supplied_metadata["datadoc"]["dataset"][field] = _cast_to_date_type(
            supplied_metadata["datadoc"]["dataset"].get(field, None),
        )
        for v in supplied_metadata["datadoc"]["variables"]:
            v[field] = _cast_to_date_type(v.get(field, None))

    supplied_metadata["datadoc"]["document_version"] = "3.3.0"
    return supplied_metadata


def handle_version_3_1_0(supplied_metadata: dict[str, Any]) -> dict[str, Any]:
    """Handle breaking changes for version 3.1.0.

    This function modifies the supplied metadata to accommodate breaking
    changes introduced in version 3.2.0. Specifically, it updates the
    'data_source' field in both the 'dataset' and 'variables' sections of the
    supplied metadata dictionary by converting value to string.
    The 'document_version' field is also updated to "3.2.0".

    Args:
        supplied_metadata: The metadata dictionary to be updated.

    Returns:
        The updated metadata dictionary.
    """
    data = supplied_metadata["datadoc"]["dataset"]["data_source"]

    if data is not None:
        supplied_metadata["datadoc"]["dataset"]["data_source"] = str(
            data[0]["languageText"],
        )

    for i in range(len(supplied_metadata["datadoc"]["variables"])):
        data = supplied_metadata["datadoc"]["variables"][i]["data_source"]
        if data is not None:
            supplied_metadata["datadoc"]["variables"][i]["data_source"] = str(
                data[0]["languageText"],
            )

    supplied_metadata["datadoc"]["document_version"] = "3.2.0"
    return supplied_metadata


def handle_version_2_2_0(supplied_metadata: dict[str, Any]) -> dict[str, Any]:
    """Handle breaking changes for version 2.2.0.

    This function modifies the supplied metadata to accommodate breaking changes
    introduced in version 3.1.0. Specifically, it updates the 'subject_field' in
    the 'dataset' section of the supplied metadata dictionary by converting it to
    a string. It also removes the 'register_uri' field from the 'dataset'.
    Additionally, it removes 'sentinel_value_uri' from each variable,
    sets 'special_value' and 'custom_type' fields to None, and updates
    language strings in the 'variables' and 'dataset' sections.
    The 'document_version' is updated to "3.1.0".

    Args:
        supplied_metadata: The metadata dictionary to be updated.

    Returns:
        The updated metadata dictionary.
    """
    if supplied_metadata["datadoc"]["dataset"]["subject_field"] is not None:
        data = supplied_metadata["datadoc"]["dataset"]["subject_field"]
        supplied_metadata["datadoc"]["dataset"]["subject_field"] = str(
            data["nb"] or data["nn"] or data["en"],
        )

    _remove_element_from_model(supplied_metadata["datadoc"]["dataset"], "register_uri")

    for i in range(len(supplied_metadata["datadoc"]["variables"])):
        _remove_element_from_model(
            supplied_metadata["datadoc"]["variables"][i],
            "sentinel_value_uri",
        )
        supplied_metadata["datadoc"]["variables"][i]["special_value"] = None
        supplied_metadata["datadoc"]["variables"][i]["custom_type"] = None
        supplied_metadata["datadoc"]["variables"][i] = (
            _find_and_update_language_strings(
                supplied_metadata["datadoc"]["variables"][i],
            )
        )
    supplied_metadata["datadoc"]["dataset"]["custom_type"] = None
    supplied_metadata["datadoc"]["dataset"] = _find_and_update_language_strings(
        supplied_metadata["datadoc"]["dataset"],
    )
    supplied_metadata["datadoc"]["document_version"] = "3.1.0"
    return supplied_metadata


def add_container(existing_metadata: dict) -> dict:
    """Add container for previous versions.

    Adds a container structure for previous versions of metadata.
    This function wraps the existing metadata in a new container structure
    that includes the 'document_version', 'datadoc', and 'pseudonymization'
    fields. The 'document_version' is set to "0.0.1" and 'pseudonymization'
    is set to None.

    Args:
        existing_metadata: The original metadata dictionary to be wrapped.

    Returns:
        A new dictionary containing the wrapped metadata with additional fields.
    """
    return {
        "document_version": "0.0.1",
        "datadoc": existing_metadata,
        "pseudonymization": None,
    }


def handle_version_2_1_0(supplied_metadata: dict[str, Any]) -> dict[str, Any]:
    """Handle breaking changes for version 2.1.0.

    This function modifies the supplied metadata to accommodate breaking changes
    introduced in version 2.2.0. Specifically, it updates the 'owner' field in
    the 'dataset' section of the supplied metadata dictionary by converting it
    from a LanguageStringType to a string.
    The 'document_version' is updated to "2.2.0".

    Args:
        supplied_metadata: The metadata dictionary to be updated.

    Returns:
        The updated metadata dictionary.
    """
    data = supplied_metadata["dataset"]["owner"]
    supplied_metadata["dataset"]["owner"] = str(data["nb"] or data["nn"] or data["en"])
    supplied_metadata["document_version"] = "2.2.0"
    return add_container(supplied_metadata)


def handle_version_1_0_0(supplied_metadata: dict[str, Any]) -> dict[str, Any]:
    """Handle breaking changes for version 1.0.0.

    This function modifies the supplied metadata to accommodate breaking changes
    introduced in version 2.1.0. Specifically, it updates the date fields
    'metadata_created_date' and 'metadata_last_updated_date' to ISO 8601 format
    with UTC timezone. It also converts the 'data_source' field from a string to a
    dictionary with language keys if necessary and removes the 'data_source_path'
    field.
    The 'document_version' is updated to "2.1.0".

    Args:
        supplied_metadata: The metadata dictionary to be updated.

    Returns:
        The updated metadata dictionary.

    """
    datetime_fields = [("metadata_created_date"), ("metadata_last_updated_date")]
    for field in datetime_fields:
        if supplied_metadata["dataset"][field]:
            supplied_metadata["dataset"][field] = datetime.isoformat(
                datetime.fromisoformat(supplied_metadata["dataset"][field]).astimezone(
                    tz=timezone.utc,
                ),
                timespec="seconds",
            )
    if isinstance(supplied_metadata["dataset"]["data_source"], str):
        supplied_metadata["dataset"]["data_source"] = {
            "en": supplied_metadata["dataset"]["data_source"],
            "nn": "",
            "nb": "",
        }

    _remove_element_from_model(supplied_metadata["dataset"], "data_source_path")

    supplied_metadata["document_version"] = "2.1.0"
    return supplied_metadata


def handle_version_0_1_1(supplied_metadata: dict[str, Any]) -> dict[str, Any]:
    """Handle breaking changes for version 0.1.1.

    This function modifies the supplied metadata to accommodate breaking changes
    introduced in version 1.0.0. Specifically, it renames certain keys within the
    `dataset` and `variables` sections, and replaces empty string values with
    `None` for `dataset` keys.

    Args:
        supplied_metadata: The metadata dictionary that needs to be updated.

    Returns:
        The updated metadata dictionary.

    References:
        PR ref: https://github.com/statisticsnorway/ssb-datadoc-model/pull/4
    """
    key_renaming = [
        ("metadata_created_date", "created_date"),
        ("metadata_created_by", "created_by"),
        ("metadata_last_updated_date", "last_updated_date"),
        ("metadata_last_updated_by", "last_updated_by"),
    ]
    for new_key, old_key in key_renaming:
        supplied_metadata["dataset"][new_key] = supplied_metadata["dataset"].pop(
            old_key,
        )
    # Replace empty strings with None, empty strings are not valid for LanguageStrings values
    supplied_metadata["dataset"] = {
        k: None if v == "" else v for k, v in supplied_metadata["dataset"].items()
    }

    key_renaming = [("data_type", "datatype")]

    for i in range(len(supplied_metadata["variables"])):
        for new_key, old_key in key_renaming:
            supplied_metadata["variables"][i][new_key] = supplied_metadata["variables"][
                i
            ].pop(
                old_key,
            )

    return supplied_metadata


# Register all the supported versions and their handlers.
# MUST be ordered from oldest to newest.
BackwardsCompatibleVersion(version="0.1.1", handler=handle_version_0_1_1)
BackwardsCompatibleVersion(version="1.0.0", handler=handle_version_1_0_0)
BackwardsCompatibleVersion(
    version="2.1.0",
    handler=handle_version_2_1_0,
)  # A container must be created at this version
BackwardsCompatibleVersion(version="2.2.0", handler=handle_version_2_2_0)
BackwardsCompatibleVersion(version="3.1.0", handler=handle_version_3_1_0)
BackwardsCompatibleVersion(version="3.2.0", handler=handle_version_3_2_0)
BackwardsCompatibleVersion(version="3.3.0", handler=handle_version_3_3_0)
BackwardsCompatibleVersion(version="4.0.0", handler=handle_version_4_0_0)
BackwardsCompatibleVersion(version="5.0.1", handler=handle_current_version)


def upgrade_metadata(fresh_metadata: dict[str, Any]) -> dict[str, Any]:
    """Upgrade the metadata to the latest version using registered handlers.

    This function checks the version of the provided metadata and applies a series
    of upgrade handlers to migrate the metadata to the latest version.
    It starts from the provided version and applies all subsequent handlers in
    sequence. If the metadata is already in the latest version or the version
    cannot be determined, appropriate actions are taken.

    Args:
        fresh_metadata: The metadata dictionary to be upgraded. This dictionary
            must include version information that determines which handlers to apply.

    Returns:
        The upgraded metadata dictionary, after applying all necessary handlers.

    Raises:
        UnknownModelVersionError: If the metadata's version is unknown or unsupported.
    """
    # Special case for current version, we expose the current_model_version parameter for test purposes
    if is_metadata_in_container_structure(fresh_metadata):
        if fresh_metadata["datadoc"] is None:
            return fresh_metadata
        supplied_version = fresh_metadata["datadoc"][VERSION_FIELD_NAME]
    else:
        supplied_version = fresh_metadata[VERSION_FIELD_NAME]
    start_running_handlers = False
    # Run all the handlers in order from the supplied version onwards
    for k, v in SUPPORTED_VERSIONS.items():
        if k == supplied_version:
            start_running_handlers = True
        if start_running_handlers:
            fresh_metadata = v.handler(fresh_metadata)
    if not start_running_handlers:
        raise UnknownModelVersionError(supplied_version)
    return fresh_metadata


def is_metadata_in_container_structure(
    metadata: dict,
) -> bool:
    """Check if the metadata is in the container structure.

    At a certain point a metadata 'container' was introduced.
    The container provides a structure for different 'types' of metadata, such as
    'datadoc', 'pseudonymization' etc.
    This function determines if the given metadata dictionary follows this container
    structure by checking for the presence of the 'datadoc' field.

    Args:
        metadata: The metadata dictionary to check.

    Returns:
        True if the metadata is in the container structure (i.e., contains the
        'datadoc' field), False otherwise.
    """
    return "datadoc" in metadata



import logging
from datetime import date
from os import PathLike
from pathlib import Path

from pydantic import ConfigDict
from pydantic import PrivateAttr

from dapla_metadata.variable_definitions._generated.vardef_client.api.draft_variable_definitions_api import (
    DraftVariableDefinitionsApi,
)
from dapla_metadata.variable_definitions._generated.vardef_client.api.patches_api import (
    PatchesApi,
)
from dapla_metadata.variable_definitions._generated.vardef_client.api.validity_periods_api import (
    ValidityPeriodsApi,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.complete_response import (
    CompleteResponse,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.patch import (
    Patch,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.update_draft import (
    UpdateDraft,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.validity_period import (
    ValidityPeriod,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.variable_status import (
    VariableStatus,
)
from dapla_metadata.variable_definitions._utils import config
from dapla_metadata.variable_definitions._utils._client import VardefClient
from dapla_metadata.variable_definitions._utils.variable_definition_files import (
    _convert_to_yaml_output,
)
from dapla_metadata.variable_definitions._utils.variable_definition_files import (
    _read_file_to_model,
)
from dapla_metadata.variable_definitions._utils.variable_definition_files import (
    create_variable_yaml,
)
from dapla_metadata.variable_definitions.exceptions import (
    publishing_blocked_error_handler,
)
from dapla_metadata.variable_definitions.exceptions import vardef_exception_handler
from dapla_metadata.variable_definitions.exceptions import vardef_file_error_handler

logger = logging.getLogger(__name__)


class VariableDefinition(CompleteResponse):
    """A Variable Definition.

    - Provides access to the fields of the specific Variable Definition.
    - Provides methods to access Patches and Validity Periods of this Variable Definition.
    - Provides methods allowing maintenance of this Variable Definition.

    Args:
        CompleteResponse: The Pydantic model superclass, representing a Variable Definition.
    """

    _file_path: Path | None = PrivateAttr(None)

    model_config = ConfigDict(use_enum_values=True, str_strip_whitespace=True)

    def get_file_path(self) -> Path | None:
        """Get the file path where the variable definition has been written to for editing."""
        return self._file_path

    def set_file_path(self, file_path: Path | None) -> None:
        """Set the file path where the variable definition has been written to for editing."""
        self._file_path = file_path

    @staticmethod
    def from_model(
        model: CompleteResponse,
    ) -> "VariableDefinition":
        """Create a VariableDefinition instance from a CompleteResponse."""
        return VariableDefinition.model_construct(**model.model_dump())

    @vardef_exception_handler
    def list_validity_periods(self) -> list["VariableDefinition"]:
        """List all Validity Periods for this Variable Definition."""
        return [
            VariableDefinition.from_model(validity_period)
            for validity_period in ValidityPeriodsApi(
                VardefClient.get_client(),
            ).list_validity_periods(
                variable_definition_id=self.id,
            )
        ]

    @vardef_exception_handler
    def list_patches(self) -> list["VariableDefinition"]:
        """List all Patches for this Variable Definition."""
        return [
            VariableDefinition.from_model(patch)
            for patch in PatchesApi(VardefClient.get_client()).list_patches(
                variable_definition_id=self.id,
            )
        ]

    @publishing_blocked_error_handler
    @vardef_exception_handler
    def update_draft(
        self,
        update_draft: UpdateDraft,
    ) -> "VariableDefinition":
        """Update this Variable Definition.

        - Variable definition must have status 'DRAFT'.
        - Supply only the fields to be changed. Other fields will retain their current values.

        Args:
            update_draft: The input with updated values.

        Returns:
            VariableDefinition: Updated Variable definition with all details.
        """
        updated = VariableDefinition.from_model(
            DraftVariableDefinitionsApi(
                VardefClient.get_client(),
            ).update_variable_definition_by_id(
                variable_definition_id=self.id,
                active_group=config.get_active_group(),
                update_draft=update_draft,
            ),
        )
        self.__dict__.update(updated)

        logger.info(
            "✅ Successfully updated variable definition '%s' with ID '%s'",
            updated.short_name,
            updated.id,
        )
        return updated

    @vardef_file_error_handler
    def update_draft_from_file(
        self,
        file_path: PathLike | None = None,
    ) -> "VariableDefinition":
        """Update this Variable Definition.

        Will automatically read the relevant file pertaining to this variable definition. Can
        be overridden by specifying the file_path parameter.

        - Variable definition must have status 'DRAFT'.
        - Supply only the fields to be changed. Other fields will retain their current values.

        Args:
            file_path: Optionally specify the path to read from.

        Returns:
            VariableDefinition: Updated Variable definition with all details.
        """
        return self.update_draft(
            _read_file_to_model(
                file_path or self.get_file_path(),
                UpdateDraft,
            ),
        )

    @vardef_exception_handler
    def delete_draft(
        self,
    ) -> str:
        """Delete this Variable definition.

        Variable definition must have status 'DRAFT'.

        Returns:
            str: A message if the operation was succsessful.

        """
        DraftVariableDefinitionsApi(
            VardefClient.get_client(),
        ).delete_variable_definition_by_id(
            variable_definition_id=self.id,
            active_group=config.get_active_group(),
        )
        return f"✅ Variable {self.id} safely deleted"

    @vardef_exception_handler
    def get_patch(self, patch_id: int) -> "VariableDefinition":
        """Get a single Patch by ID.

        Args:
            patch_id (int): The ID of the patch.

        Returns:
            VariableDefinition: The desired patch.
        """
        return VariableDefinition.from_model(
            PatchesApi(VardefClient.get_client()).get_patch(
                variable_definition_id=self.id,
                patch_id=patch_id,
            ),
        )

    @vardef_exception_handler
    def create_patch(
        self,
        patch: Patch,
        valid_from: date | None = None,
    ) -> "VariableDefinition":
        """Create a new Patch for this Variable Definition.

        Patches are to be used for minor changes which don't require a new Validity Period.
        Examples of reasons for creating a new Patch:
        - Correcting a typo
        - Adding a translation
        - Adding a subject field

        Supply only the fields to be changed. Other fields will retain their current values.

        Args:
            patch: The input for a new patch.
            valid_from: Optional date for selecting a Validity Period to create patch in. The date must
                        exactly match the Validity Period `valid_from`. If value is None the patch is
                        created in the last validity period.

        Returns:
            VariableDefinition: Variable Definition with all details.

        """
        new_patch = VariableDefinition.from_model(
            PatchesApi(
                VardefClient.get_client(),
            ).create_patch(
                variable_definition_id=self.id,
                active_group=config.get_active_group(),
                patch=patch,
                valid_from=valid_from,
            ),
        )
        self.__dict__.update(new_patch)

        logger.info(
            "✅ Successfully created patch with patch ID '%s' for variable definition '%s' with ID '%s'",
            new_patch.patch_id,
            new_patch.short_name,
            new_patch.id,
        )
        return new_patch

    @vardef_file_error_handler
    def create_patch_from_file(
        self,
        file_path: PathLike | None = None,
        valid_from: date | None = None,
    ) -> "VariableDefinition":
        """Create a new Patch for this Variable Definition from a file.

        Will automatically read the relevant file pertaining to this variable definition. Can
        be overridden by specifying the file_path parameter.

        Patches are to be used for minor changes which don't require a new Validity Period.
        Examples of reasons for creating a new Patch:
          - Correcting a typo
          - Adding a translation
          - Adding a subject field

        Supply only the fields to be changed. Other fields will retain their current values.

        Args:
            file_path: Optionally specify the path to read from.
            valid_from: Optional date for selecting a Validity Period to create patch in. The date must
                        exactly match the Validity Period `valid_from`. If value is None the patch is
                        created in the last validity period.

        Returns:
            VariableDefinition: Variable Definition with all details.
        """
        return self.create_patch(
            patch=_read_file_to_model(
                file_path or self.get_file_path(),
                Patch,
            ),
            valid_from=valid_from,
        )

    @vardef_exception_handler
    def create_validity_period(
        self,
        validity_period: ValidityPeriod,
    ) -> "VariableDefinition":
        """Create a new Validity Period for this Variable Definition.

        In order to create a new Validity Period input must contain updated
        'definition' text for all present languages and a new valid from.

        A new Validity Period should be created only when the fundamental definition
        of the variable has changed. This way the previous definition can be preserved
        for use in historical data.

        Args:
            validity_period: The input for new Validity Period

        Returns:
            VariableDefinition: Variable Definition with all details.
        """
        new_validity_period = VariableDefinition.from_model(
            ValidityPeriodsApi(
                VardefClient.get_client(),
            ).create_validity_period(
                variable_definition_id=self.id,
                active_group=config.get_active_group(),
                validity_period=validity_period,
            ),
        )
        self.__dict__.update(new_validity_period)

        logger.info(
            "✅ Successfully created validity period that is valid from '%s' for variable definition '%s' with ID '%s'",
            new_validity_period.valid_from,
            new_validity_period.short_name,
            new_validity_period.id,
        )
        return new_validity_period

    @vardef_file_error_handler
    def create_validity_period_from_file(
        self,
        file_path: PathLike | None = None,
    ) -> "VariableDefinition":
        """Create a new ValidityPeriod for this Variable Definition from a file.

        In order to create a new Validity Period the input file must contain updated
        'definition' text for all present languages and a new valid from.

        Args:
            file_path: Optionally specify the path to read from.

        Returns:
            VariableDefinition: Variable Definition with all details.
        """
        return self.create_validity_period(
            validity_period=_read_file_to_model(
                file_path or self.get_file_path(),
                ValidityPeriod,
            ),
        )

    @publishing_blocked_error_handler
    def publish_internal(self) -> "VariableDefinition":
        """Publish this variable definition internally."""
        if self.variable_status != VariableStatus.DRAFT.name:
            msg = "That won't work here. Only variable definitions with status DRAFT may be published internally."
            raise ValueError(
                msg,
            )
        update = self.update_draft(
            UpdateDraft(variable_status=VariableStatus.PUBLISHED_INTERNAL),
        )
        logger.info(
            "✅ Variable definition '%s' with ID '%s' successfully published, new status: %s",
            update.short_name,
            update.id,
            update.variable_status,
        )
        return update

    @publishing_blocked_error_handler
    def publish_external(self) -> "VariableDefinition":
        """Publish this variable definition externally."""
        if self.variable_status == VariableStatus.PUBLISHED_EXTERNAL.name:
            msg = "That won't work here. The variable definition is already published."
            raise ValueError(
                msg,
            )
        if self.variable_status is VariableStatus.DRAFT:
            update = self.update_draft(
                UpdateDraft(variable_status=VariableStatus.PUBLISHED_EXTERNAL),
            )
        else:
            update = self.create_patch(
                Patch(variable_status=VariableStatus.PUBLISHED_EXTERNAL),
            )
        logger.info(
            "✅ Variable definition '%s' with ID '%s' successfully published, new status: %s",
            update.short_name,
            update.id,
            update.variable_status,
        )
        return update

    def to_file(self) -> "VariableDefinition":
        """Write this variable definition to file."""
        file_path = create_variable_yaml(
            model_instance=self,
        )
        self.set_file_path(file_path)
        logger.info(
            f"✅ Created editable variable definition file at {file_path}",  # noqa: G004
        )
        return self

    def to_dict(self) -> dict:
        """Return as dictionary."""
        return super().to_dict()

    def __str__(self) -> str:
        """Format as indented YAML."""
        return _convert_to_yaml_output(self)

    def __repr__(self) -> str:
        """Format as indented YAML."""
        return _convert_to_yaml_output(self)



"""Extract info from a path following SSB's dataset naming convention."""

from __future__ import annotations

import logging
import pathlib
import re
from abc import ABC
from abc import abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING
from typing import Final
from typing import Literal

import arrow
from cloudpathlib import GSPath
from datadoc_model.all_optional.model import DataSetState

if TYPE_CHECKING:
    import datetime
    import os
    from datetime import date

logger = logging.getLogger(__name__)

GS_PREFIX_FROM_PATHLIB = "gs:/"


@dataclass
class DateFormat(ABC):
    """A super class for date formats."""

    name: str
    regex_pattern: str
    arrow_pattern: str
    timeframe: Literal["year", "month", "day", "week"]

    @abstractmethod
    def get_floor(self, period_string: str) -> date | None:
        """Abstract method implemented in the child class.

        Return the first date of the timeframe period.

        Args:
            period_string: A string representing the timeframe period.
        """

    @abstractmethod
    def get_ceil(self, period_string: str) -> date | None:
        """Abstract method implemented in the child class.

        Return the last date of the timeframe period.

        Args:
            period_string: A string representing the timeframe period.
        """


@dataclass
class IsoDateFormat(DateFormat):
    """A subclass of Dateformat with relevant patterns for ISO dates."""

    def get_floor(self, period_string: str) -> date | None:
        """Return first date of timeframe period defined in ISO date format.

        Examples:
            >>> ISO_YEAR_MONTH.get_floor("1980-08")
            datetime.date(1980, 8, 1)

            >>> ISO_YEAR.get_floor("2021")
            datetime.date(2021, 1, 1)
        """
        return arrow.get(period_string, self.arrow_pattern).floor(self.timeframe).date()

    def get_ceil(self, period_string: str) -> date | None:
        """Return last date of timeframe period defined in ISO date format.

        Examples:
            >>> ISO_YEAR.get_ceil("1921")
            datetime.date(1921, 12, 31)

            >>> ISO_YEAR_MONTH.get_ceil("2021-05")
            datetime.date(2021, 5, 31)
        """
        return arrow.get(period_string, self.arrow_pattern).ceil(self.timeframe).date()


ISO_YEAR = IsoDateFormat(
    name="ISO_YEAR",
    regex_pattern=r"^\d{4}$",
    arrow_pattern="YYYY",
    timeframe="year",
)
ISO_YEAR_MONTH = IsoDateFormat(
    name="ISO_YEAR_MONTH",
    regex_pattern=r"^\d{4}\-\d{2}$",
    arrow_pattern="YYYY-MM",
    timeframe="month",
)
ISO_YEAR_MONTH_DAY = IsoDateFormat(
    name="ISO_YEAR_MONTH_DAY",
    regex_pattern=r"^\d{4}\-\d{2}\-\d{2}$",
    arrow_pattern="YYYY-MM-DD",
    timeframe="day",
)
ISO_YEAR_WEEK = IsoDateFormat(
    name="ISO_YEAR_WEEK",
    regex_pattern=r"^\d{4}\-{0,1}W\d{2}$",
    arrow_pattern="W",
    timeframe="week",
)


@dataclass
class SsbDateFormat(DateFormat):
    """A subclass of Dateformat with relevant patterns for SSB unique dates.

    Attributes:
        ssb_dates: A dictionary where keys are date format strings and values
                    are corresponding date patterns specific to SSB.
    """

    ssb_dates: dict

    def get_floor(self, period_string: str) -> date | None:
        """Return first date of the timeframe period defined in SSB date format.

        Convert SSB format to date-string and return the first date.

        Args:
            period_string: A string representing the timeframe period in
                SSB format.

        Returns:
            The first date of the period if the period_string is a valid
            SSB format, otherwise None.

        Example:
            >>> SSB_BIMESTER.get_floor("2003B8")
            None

            >>> SSB_BIMESTER.get_floor("2003B4")
            datetime.date(2003, 7, 1)

            >>> SSB_BIMESTER.get_floor("2003-B4")
            datetime.date(2003, 7, 1)
        """
        try:
            year = period_string[:4]
            month = self.ssb_dates[period_string[-2:]]["start"]
            period = year + month
            return arrow.get(period, self.arrow_pattern).floor(self.timeframe).date()
        except KeyError:
            logger.exception("Error while converting to SSB date format")
            return None

    def get_ceil(self, period_string: str) -> date | None:
        """Return last date of the timeframe period defined in SSB date format.

        Convert SSB format to date-string and return the last date.

        Args:
            period_string: A string representing the timeframe period in SSB
                            format.

        Returns:
            The last date of the period if the period_string is a valid SSB format,
            otherwise None.

        Example:
            >>> SSB_TRIANNUAL.get_ceil("1999T11")
            None

            >>> SSB_HALF_YEAR.get_ceil("2024H1")
            datetime.date(2024, 6, 30)

            >>> SSB_HALF_YEAR.get_ceil("2024-H1")
            datetime.date(2024, 6, 30)
        """
        try:
            year = period_string[:4]
            month = self.ssb_dates[period_string[-2:]]["end"]
            period = year + month
            return arrow.get(period, self.arrow_pattern).ceil(self.timeframe).date()
        except KeyError:
            return None


SSB_BIMESTER = SsbDateFormat(
    name="SSB_BIMESTER",
    regex_pattern=r"^\d{4}-?[B]\d{1}$",
    arrow_pattern="YYYYMM",
    timeframe="month",
    ssb_dates={
        "B1": {
            "start": "01",
            "end": "02",
        },
        "B2": {
            "start": "03",
            "end": "04",
        },
        "B3": {
            "start": "05",
            "end": "06",
        },
        "B4": {
            "start": "07",
            "end": "08",
        },
        "B5": {
            "start": "09",
            "end": "10",
        },
        "B6": {
            "start": "11",
            "end": "12",
        },
    },
)

SSB_QUARTERLY = SsbDateFormat(
    name="SSB_QUARTERLY",
    regex_pattern=r"^\d{4}-?[Q]\d{1}$",
    arrow_pattern="YYYYMM",
    timeframe="month",
    ssb_dates={
        "Q1": {
            "start": "01",
            "end": "03",
        },
        "Q2": {
            "start": "04",
            "end": "06",
        },
        "Q3": {
            "start": "07",
            "end": "09",
        },
        "Q4": {
            "start": "10",
            "end": "12",
        },
    },
)

SSB_TRIANNUAL = SsbDateFormat(
    name="SSB_TRIANNUAL",
    regex_pattern=r"^\d{4}-?[T]\d{1}$",
    arrow_pattern="YYYYMM",
    timeframe="month",
    ssb_dates={
        "T1": {
            "start": "01",
            "end": "04",
        },
        "T2": {
            "start": "05",
            "end": "08",
        },
        "T3": {
            "start": "09",
            "end": "12",
        },
    },
)
SSB_HALF_YEAR = SsbDateFormat(
    name="SSB_HALF_YEAR",
    regex_pattern=r"^\d{4}-?[H]\d{1}$",
    arrow_pattern="YYYYMM",
    timeframe="month",
    ssb_dates={
        "H1": {
            "start": "01",
            "end": "06",
        },
        "H2": {
            "start": "07",
            "end": "12",
        },
    },
)

SUPPORTED_DATE_FORMATS: list[IsoDateFormat | SsbDateFormat] = [
    ISO_YEAR,
    ISO_YEAR_MONTH,
    ISO_YEAR_MONTH_DAY,
    ISO_YEAR_WEEK,
    SSB_BIMESTER,
    SSB_QUARTERLY,
    SSB_TRIANNUAL,
    SSB_HALF_YEAR,
]


def categorize_period_string(period: str) -> IsoDateFormat | SsbDateFormat:
    """Categorize a period string into one of the supported date formats.

    Args:
        period: A string representing the period to be categorized.

    Returns:
        An instance of either IsoDateFormat or SsbDateFormat depending on the
        format of the input period string.

    Raises:
        NotImplementedError: If the period string is not recognized as either an
            ISO or SSB date format.

    Examples:
        >>> date_format = categorize_period_string('2022-W01')
        >>> date_format.name
        ISO_YEAR_WEEK

        >>> date_format = categorize_period_string('1954T2')
        >>> date_format.name
        SSB_TRIANNUAL

        >>> categorize_period_string('unknown format')
        Traceback (most recent call last):
        ...
        NotImplementedError: Period format unknown format is not supported
    """
    for date_format in SUPPORTED_DATE_FORMATS:
        if re.match(date_format.regex_pattern, period):
            return date_format

    msg = f"Period format {period} is not supported"
    raise NotImplementedError(
        msg,
    )


class DaplaDatasetPathInfo:
    """Extract info from a path following SSB's dataset naming convention."""

    def __init__(self, dataset_path: str | os.PathLike[str]) -> None:
        """Digest the path so that it's ready for further parsing."""
        self.dataset_string = str(dataset_path)
        self.dataset_path = pathlib.Path(dataset_path)
        self.dataset_name_sections = self.dataset_path.stem.split("_")
        self._period_strings = self._extract_period_strings(self.dataset_name_sections)

    @staticmethod
    def _get_period_string_indices(dataset_name_sections: list[str]) -> list[int]:
        """Get all the indices at which period strings are found in list.

        Args:
            dataset_name_sections: A list of strings representing sections of a
                dataset name.

        Returns:
            A list of indices where period strings are found within the
            dataset_name_sections.

        Examples:
            >>> DaplaDatasetPathInfo._get_period_string_indices(['kommune', 'p2022', 'v1'])
            [1]

            >>> DaplaDatasetPathInfo._get_period_string_indices(['kommune', 'p2022-01', 'p2023-06', 'v1'])
            [1, 2]

            >>> DaplaDatasetPathInfo._get_period_string_indices(['kommune', 'p1990Q1', 'v1'])
            [1]

            >>> DaplaDatasetPathInfo._get_period_string_indices(['varehandel','v1'])
            []
        """

        def insert_p(regex: str) -> str:
            r"""Insert a 'p' as the second character.

            Args:
                regex: A string representing the regular expression pattern to be
                    modified.

            Returns:
                The modified regular expression pattern with 'p' inserted as the
                second character.

            Examples:
                >>> insert_p(r"^\d{4}[H]\d{1}$")
                '^p\d{4}[H]\d{1}$'
            """
            return regex[:1] + "p" + regex[1:]

        return [
            i
            for i, x in enumerate(dataset_name_sections)
            if any(
                re.match(insert_p(date_format.regex_pattern), x)
                for date_format in SUPPORTED_DATE_FORMATS
            )
        ]

    @staticmethod
    def _extract_period_strings(dataset_name_sections: list[str]) -> list[str]:
        """Extract period strings from dataset name sections.

        Iterates over the dataset name sections and returns a list of strings
        that match the year regex, stripping the first character. This extracts
        the year periods from the dataset name.

        Args:
            dataset_name_sections: A list of strings representing sections of a
                dataset name.

        Returns:
            A list of extracted period strings, with the first character stripped
            from each match.

        Examples:
            >>> DaplaDatasetPathInfo._extract_period_strings(['p2022', 'kommune', 'v1'])
            ['2022']

            >>> DaplaDatasetPathInfo._extract_period_strings(['p2022-01', 'p2023-06', 'kommune', 'v1'])
            ['2022-01', '2023-06']

            >>> DaplaDatasetPathInfo._extract_period_strings(['p1990Q1', 'kommune', 'v1'])
            ['1990Q1']

            >>> DaplaDatasetPathInfo._extract_period_strings(['p1990-Q1', 'kommune', 'v1'])
            ['1990-Q1']

            >>> DaplaDatasetPathInfo._extract_period_strings(['varehandel','v1'])
            []
        """
        return [
            dataset_name_sections[i][1:]
            for i in DaplaDatasetPathInfo._get_period_string_indices(
                dataset_name_sections,
            )
        ]

    def _extract_period_string_from_index(self, index: int) -> str | None:
        """Extract a period string by its index from the list of period strings.

        Args:
            index: The index of the period string to extract.

        Returns:
            The extracted period string if it exists, otherwise None.
        """
        try:
            return self._period_strings[index]
        except IndexError:
            return None

    def _extract_norwegian_dataset_state_path_part(
        self,
        dataset_state: DataSetState,
    ) -> set:
        """Extract the Norwegian dataset state path part.

        Args:
            dataset_state: The dataset state.

        Returns:
            A set of variations of the Norwegian dataset state path part.
        """
        norwegian_mappings = {
            "SOURCE_DATA": "kildedata",
            "INPUT_DATA": "inndata",
            "PROCESSED_DATA": "klargjorte_data",
            "STATISTICS": "statistikk",
            "OUTPUT_DATA": "utdata",
        }
        norwegian_state = norwegian_mappings.get(dataset_state.name)
        if norwegian_state:
            state_name = norwegian_state.lower().replace("_", " ")
            return {state_name.replace(" ", "-"), state_name.replace(" ", "_")}
        return set()

    @property
    def bucket_name(
        self,
    ) -> str | None:
        """Extract the bucket name from the dataset path.

        Returns:
            The bucket name or None if the dataset path is not a GCS path nor ssb bucketeer path.

        Examples:
            >>> DaplaDatasetPathInfo('gs://ssb-staging-dapla-felles-data-delt/datadoc/utdata/person_data_p2021_v2.parquet').bucket_name
            ssb-staging-dapla-felles-data-delt

            >>> DaplaDatasetPathInfo(pathlib.Path('gs://ssb-staging-dapla-felles-data-delt/datadoc/utdata/person_data_p2021_v2.parquet')).bucket_name
            ssb-staging-dapla-felles-data-delt

            >>> DaplaDatasetPathInfo('gs:/ssb-staging-dapla-felles-data-delt/datadoc/utdata/person_data_p2021_v2.parquet').bucket_name
            ssb-staging-dapla-felles-data-delt

            >>> DaplaDatasetPathInfo('ssb-staging-dapla-felles-data-delt/datadoc/utdata/person_data_p2021_v2.parquet').bucket_name
            None

            >>> DaplaDatasetPathInfo('ssb-staging-dapla-felles-data-delt/datadoc/utdata/person_data_p2021_v2.parquet').bucket_name
            None

            >>> DaplaDatasetPathInfo('buckets/ssb-staging-dapla-felles-data-delt/stat/utdata/person_data_p2021_v2.parquet').bucket_name
            ssb-staging-dapla-felles-data-delt

            >>> DaplaDatasetPathInfo('buckets/ssb-staging-dapla-felles-data-delt/person_data_p2021_v2.parquet').bucket_name
            ssb-staging-dapla-felles-data-delt

            >>> DaplaDatasetPathInfo('home/work/buckets/ssb-staging-dapla-felles-produkt/stat/utdata/person_data_p2021_v2.parquet').bucket_name
            ssb-staging-dapla-felles-produkt
        """
        prefix: str | None = None
        dataset_string = str(self.dataset_string)
        if GSPath.cloud_prefix in self.dataset_string:
            prefix = GSPath.cloud_prefix
            _, bucket_and_rest = dataset_string.split(prefix, 1)
        elif GS_PREFIX_FROM_PATHLIB in self.dataset_string:
            prefix = GS_PREFIX_FROM_PATHLIB
            _, bucket_and_rest = self.dataset_string.split(prefix, 1)
        elif "buckets/" in self.dataset_string:
            prefix = "buckets/"
            _, bucket_and_rest = self.dataset_string.split(prefix, 1)
        else:
            return None

        return pathlib.Path(
            bucket_and_rest,
        ).parts[0]

    @property
    def dataset_short_name(
        self,
    ) -> str | None:
        """Extract the dataset short name from the filepath.

        The dataset short name is defined as the first section of the stem, up to
        the period information or the version information if no period information
        is present.

        Returns:
            The extracted dataset short name if it can be determined, otherwise
            None.

        Examples:
            >>> DaplaDatasetPathInfo('prosjekt/befolkning/klargjorte_data/person_data_v1.parquet').dataset_short_name
            person_data

            >>> DaplaDatasetPathInfo('befolkning/inndata/sykepenger_p2022Q1_p2022Q2_v23.parquet').dataset_short_name
            sykepenger

            >>> DaplaDatasetPathInfo('my_data/simple_dataset_name.parquet').dataset_short_name
            simple_dataset_name

            >>> DaplaDatasetPathInfo('gs:/ssb-staging-dapla-felles-data-delt/datadoc/utdata/person_data_p2021_v2.parquet').dataset_short_name
            person_data

            >>> DaplaDatasetPathInfo('buckets/ssb-staging-dapla-felles-data-delt/stat/utdata/folk_data_p2021_v2.parquet').dataset_short_name
            folk_data

            >>> DaplaDatasetPathInfo('buckets/ssb-staging-dapla-felles-data-delt/stat/utdata/dapla/bus_p2021_v2.parquet').dataset_short_name
            bus
        """
        if self.contains_data_from or self.contains_data_until:
            short_name_sections = self.dataset_name_sections[
                : min(
                    DaplaDatasetPathInfo._get_period_string_indices(
                        self.dataset_name_sections,
                    ),
                )
            ]
        elif self.dataset_version:
            short_name_sections = self.dataset_name_sections[:-1]
        else:
            short_name_sections = self.dataset_name_sections

        return "_".join(short_name_sections)

    @property
    def contains_data_from(self) -> datetime.date | None:
        """The earliest date from which data in the dataset is relevant for.

        Returns:
            The earliest relevant date for the dataset if available, otherwise None.
        """
        period_string = self._extract_period_string_from_index(0)
        if not period_string or (
            len(self._period_strings) > 1 and period_string > self._period_strings[1]
        ):
            return None
        date_format = categorize_period_string(period_string)
        return date_format.get_floor(period_string)

    @property
    def contains_data_until(self) -> datetime.date | None:
        """The latest date until which data in the dataset is relevant for.

        Returns:
            The latest relevant date for the dataset if available, otherwise None.
        """
        first_period_string = self._extract_period_string_from_index(0)
        second_period_string = self._extract_period_string_from_index(1)
        period_string = second_period_string or first_period_string
        if not period_string or (
            second_period_string
            and first_period_string is not None
            and second_period_string < first_period_string
        ):
            return None
        date_format = categorize_period_string(period_string)
        return date_format.get_ceil(period_string)

    @property
    def dataset_state(
        self,
    ) -> DataSetState | None:
        """Extract the dataset state from the path.

        We assume that files are saved in the Norwegian language as specified by
        SSB.

        Returns:
            The extracted dataset state if it can be determined from the path,
            otherwise None.

        Examples:
            >>> DaplaDatasetPathInfo('klargjorte_data/person_data_v1.parquet').dataset_state
            <DataSetState.PROCESSED_DATA: 'PROCESSED_DATA'>

            >>> DaplaDatasetPathInfo('klargjorte-data/person_data_v1.parquet').dataset_state
            <DataSetState.PROCESSED_DATA: 'PROCESSED_DATA'>

            >>> DaplaDatasetPathInfo('utdata/min_statistikk/person_data_v1.parquet').dataset_state
            <DataSetState.OUTPUT_DATA: 'OUTPUT_DATA'>

            >>> DaplaDatasetPathInfo('buckets/bucket_name/stat_name/inndata/min_statistikk/person_data_v1.parquet').dataset_state
            <DataSetState.INPUT_DATA: 'INPUT_DATA'>

            >>> DaplaDatasetPathInfo('my_special_data/person_data_v1.parquet').dataset_state
            None
        """
        dataset_path_parts = set(self.dataset_path.parts)
        for state in DataSetState:
            norwegian_variations = self._extract_norwegian_dataset_state_path_part(
                state,
            )
            if norwegian_variations.intersection(dataset_path_parts):
                return state
        return None

    @property
    def dataset_version(
        self,
    ) -> str | None:
        """Extract version information if exists in filename.

        Returns:
            The extracted version information if available in the filename,
            otherwise None.

        Examples:
            >>> DaplaDatasetPathInfo('person_data_v1.parquet').dataset_version
            '1'

            >>> DaplaDatasetPathInfo('person_data_v20.parquet').dataset_version
            '20'

            >>> DaplaDatasetPathInfo('person_data.parquet').dataset_version
            None

            >>> DaplaDatasetPathInfo('buckets/bucket_name/stat_name/inndata/min_statistikk/person_data_v1.parquet').dataset_version
            '1'

            >>> DaplaDatasetPathInfo('buckets/bucket_name/stat_name/inndata/min_statistikk/person_data.parquet').dataset_version
            None
        """
        minimum_elements_in_file_name: Final[int] = 2
        minimum_characters_in_version_string: Final[int] = 2
        if len(self.dataset_name_sections) >= minimum_elements_in_file_name:
            last_filename_element = str(self.dataset_name_sections[-1])
            if (
                len(last_filename_element) >= minimum_characters_in_version_string
                and last_filename_element[0:1] == "v"
                and last_filename_element[1:].isdigit()
            ):
                return last_filename_element[1:]
        return None

    def _get_left_parts(
        self,
        dataset_path_parts: list[str],
        state_index: int,
    ) -> list[str]:
        """Retrieve the path parts before the dataset state, considering bucket prefixes."""
        bucket_prefix = {"gs:", "buckets"}
        left_parts = dataset_path_parts[:state_index]

        # Stop checking beyond the bucket prefix
        prefix_intersection = bucket_prefix & set(left_parts)
        if prefix_intersection:
            first_prefix = min(
                left_parts.index(prefix) for prefix in prefix_intersection
            )
            left_parts = left_parts[first_prefix:]

        return (
            []
            if left_parts == ["/"]
            or (left_parts[0] in bucket_prefix and len(left_parts) <= 2)
            else left_parts
        )

    @property
    def statistic_short_name(
        self,
    ) -> str | None:
        """Extract the statistical short name from the filepath.

        Extract the statistical short name from the filepath either after bucket name or right before the
        dataset state based on the Dapla filepath naming convention.

        Returns:
            The extracted statistical short name if it can be determined,
            otherwise None.

        Examples:
            >>> DaplaDatasetPathInfo('prosjekt/befolkning/klargjorte_data/person_data_v1.parquet').statistic_short_name
            befolkning

            >>> DaplaDatasetPathInfo('buckets/prosjekt/befolkning/person_data_v1.parquet').statistic_short_name
            befolkning

            >>> DaplaDatasetPathInfo('befolkning/inndata/person_data_v1.parquet').statistic_short_name
            befolkning

            >>> DaplaDatasetPathInfo('buckets/bucket_name/stat_name/inndata/min_statistikk/person_data.parquet').statistic_short_name
            stat_name

            >>> DaplaDatasetPathInfo('buckets/stat_name/utdata/person_data.parquet').statistic_short_name
            None

            >>> DaplaDatasetPathInfo('befolkning/person_data.parquet').statistic_short_name
            None

            >>> DaplaDatasetPathInfo('buckets/produkt/befolkning/utdata/person_data.parquet').statistic_short_name
            befolkning

            >>> DaplaDatasetPathInfo('resources/buckets/produkt/befolkning/utdata/person_data.parquet').statistic_short_name
            befolkning

            >>> DaplaDatasetPathInfo('gs://statistikk/produkt/klargjorte-data/persondata_p1990-Q1_p2023-Q4_v1/aar=2019/data.parquet').statistic_short_name
            produkt

            >>> DaplaDatasetPathInfo('gs://statistikk/produkt/persondata_p1990-Q1_p2023-Q4_v1/aar=2019/data.parquet').statistic_short_name
            None

            >>> DaplaDatasetPathInfo('buckets/ssb-staging-dapla-felles-data-delt/person_data_p2021_v2.parquet').statistic_short_name
            None
        """
        if not self.dataset_state:
            if self.bucket_name:
                parts = self.dataset_path.parent.parts

                if self.bucket_name not in parts:
                    return None

                # Find the index of bucket_name in the path
                bucket_name_index = self.dataset_path.parent.parts.index(
                    self.bucket_name,
                )

                # If there are parts after bucket_name, return the part immediately after it
                if len(self.dataset_path.parent.parts) > bucket_name_index + 1:
                    return self.dataset_path.parent.parts[bucket_name_index + 1]

            return None

        dataset_state_names = self._extract_norwegian_dataset_state_path_part(
            self.dataset_state,
        )
        dataset_path_parts = list(self.dataset_path.parts)

        for state in dataset_state_names:
            if state not in dataset_path_parts:
                continue

            index = dataset_path_parts.index(state)

            if index == 0:
                continue

            left_parts = self._get_left_parts(dataset_path_parts, index)

            if not left_parts:
                return None

            return dataset_path_parts[index - 1]

        return None

    def path_complies_with_naming_standard(self) -> bool:
        """Check if path is valid according to SSB standard.

        Read more about SSB naming convention in the Dapla manual:
        https://manual.dapla.ssb.no/statistikkere/navnestandard.html

        Returns:
            True if the path conforms to the SSB naming standard, otherwise False.
        """
        return bool(
            self.dataset_state
            and self.statistic_short_name
            and self.contains_data_from
            and self.contains_data_until
            and self.dataset_version,
        )



"""Constants for variable definitions."""

from datetime import date

VARIABLE_DEFINITIONS_DIR = "variable_definitions"

VARIABLE_STATUS_FIELD_NAME = "variable_status"
OWNER_FIELD_NAME = "owner"

TEMPLATE_HEADER = "--- Variabeldefinisjon mal ---\n"
HEADER = "--- Variabeldefinisjon ---\n"

TEMPLATE_SECTION_HEADER_STATUS = "\n--- Statusfelt. Verdi 'DRAFT' før publisering. Ikke rediger hvis du oppretter en ny variabeldefinisjon. ---\n"
TEMPLATE_SECTION_HEADER_OWNER = "\n--- Eierteam og grupper. Ikke rediger hvis du oppretter en ny variabeldefinisjon, verdien genereres ---\n"
TEMPLATE_SECTION_HEADER_MACHINE_GENERATED = (
    "\n--- Maskin-genererte felt. Ikke rediger. ---\n"
)

TEMPLATE_HEADER_EN = "--- Variable definition template ---\n"
HEADER_EN = "--- Variable definition ---\n"
TEMPLATE_SECTION_HEADER_STATUS_EN = "\n--- Status field. Value 'DRAFT' before publishing. Do not edit if creating new variable defintion ---\n"
TEMPLATE_SECTION_HEADER_OWNER_EN = "\n--- Owner team and groups. Do not edit if creating new variable defintion, value is generated ---\n"
TEMPLATE_SECTION_HEADER_MACHINE_GENERATED_EN = (
    "\n--- Machine generated fields. Do not edit ---\n"
)

DEFAULT_DATE = date(1000, 1, 1)

MACHINE_GENERATED_FIELDS = [
    "id",
    "patch_id",
    "created_at",
    "created_by",
    "last_updated_at",
    "last_updated_by",
]

OPTIONAL_FIELD = "~ Valgfritt felt ~"
REQUIRED_FIELD = "! Obligatorisk felt !"

YAML_STR_TAG = "tag:yaml.org,2002:str"

BLOCK_FIELDS = [
    "definition",
    "name",
    "contact.title",
    "comment",
]

DOUBLE_QUOTE_FIELDS = [
    "unit_types",
    "subject_fields",
    "related_variable_definition_uris",
    "owner",
    "short_name",
    "classification_reference",
    "measurement_type",
    "external_reference_uri",
    "created_by",
    "id",
    "last_updated_by",
]

PUBLISHING_BLOCKED_ERROR_MESSAGE = "Publishing blocked: Publishing variable definitions is not allowed until further notice."
VARDEF_PROD_URL = "https://metadata.intern.ssb.no"
VARDEF_TEST_URL = "https://metadata.intern.test.ssb.no"



"""Vardef client exceptions."""

import json
from functools import wraps
from http import HTTPStatus
from types import MappingProxyType

import urllib3
from pytz import UnknownTimeZoneError
from ruamel.yaml.error import YAMLError

from dapla_metadata.variable_definitions._generated.vardef_client.exceptions import (
    OpenApiException,
)
from dapla_metadata.variable_definitions._generated.vardef_client.exceptions import (
    UnauthorizedException,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.variable_status import (
    VariableStatus,
)
from dapla_metadata.variable_definitions._utils._client import VardefClient
from dapla_metadata.variable_definitions._utils.constants import (
    PUBLISHING_BLOCKED_ERROR_MESSAGE,
)
from dapla_metadata.variable_definitions._utils.constants import VARDEF_PROD_URL

# Use MappingProxyType so the dict is immutable
STATUS_EXPLANATIONS: MappingProxyType[HTTPStatus | None, str] = MappingProxyType(
    {
        HTTPStatus.BAD_REQUEST: "There was a problem with the supplied data. Please review the data you supplied based on the details below.",
        HTTPStatus.UNAUTHORIZED: "There is a problem with your token, it may have expired. Try logging out, logging in and trying again.",
        HTTPStatus.FORBIDDEN: "Forbidden. You don't have access to this, possibly because your team is not an owner of this variable definition.",
        HTTPStatus.NOT_FOUND: "The variable definition was not found, check that the `short_name` or `id` is correct and that the variable exists.",
        HTTPStatus.METHOD_NOT_ALLOWED: "That won't work here. The status of your variable definition likely doesn't allow for this to happen.",
        HTTPStatus.CONFLICT: "There was a conflict with existing data. Please change your data and try again.",
        None: "",
    },
)


class VardefClientError(Exception):
    """Custom exception to represent errors encountered in the Vardef client.

    This exception extracts and formats error details from a JSON response body
    provided by the Vardef API, enabling more descriptive error messages.
    If the response body cannot be parsed as JSON or lacks expected keys,
    default values are used to provide meaningful feedback.
    """

    def __init__(self, response_body: str) -> None:
        """Initialize the exception with a JSON-formatted response body.

        Args:
            response_body (str): The JSON string containing error details
                                from the Vardef API response.

        Attributes:
            status (str): The status code from the response, or "Unknown status"
                        if not available or the response is invalid.
            detail (str || list): A detailed error message from the response, or
                        "No detail provided" if not provided. If "Constraint violation"
                        detail contains 'field' and 'message'.
            response_body (str): The raw response body string, stored for
                                debugging purposes.
        """
        self.status: int | None = None
        self.detail: str = ""
        try:
            data = json.loads(response_body)
            self.status = data.get("status")
            if data.get("title") == "Constraint Violation":
                violations = data.get("violations", [])
                self.detail = "\n" + "\n".join(
                    f"{violation.get('field', 'Unknown field')}: {violation.get('message', 'No message provided')}"
                    for violation in violations
                )
            else:
                self.detail = data.get("detail")

            if self.detail:
                self.detail = f"\nDetail: {self.detail}"
            self.response_body = response_body
        except (json.JSONDecodeError, TypeError):
            self.detail = "Could not decode error response from API"

        super().__init__(
            f"{self._get_status_explanation(self.status)}{self.detail if self.detail else ''}",
        )

    @staticmethod
    def _get_status_explanation(status: int | None) -> str:
        return (
            ""
            if not status
            else (STATUS_EXPLANATIONS.get(HTTPStatus(status)) or f"Status {status}:")
        )


def vardef_exception_handler(method):  # noqa: ANN201, ANN001
    """Decorator for handling exceptions in Vardef."""

    @wraps(method)
    def _impl(self, *method_args, **method_kwargs):  # noqa: ANN001, ANN002, ANN003
        try:
            return method(self, *method_args, **method_kwargs)
        except urllib3.exceptions.HTTPError as e:
            # Catch all urllib3 exceptions by catching the base class.
            # These exceptions typically arise from lower level network problems.
            raise VardefClientError(
                json.dumps(
                    {
                        "status": None,
                        "title": "Network problems",
                        "detail": f"""There was a network problem when sending the request to the server. Try again shortly.
Original exception:
{getattr(e, "message", repr(e))}""",
                    },
                ),
            ) from e
        except UnauthorizedException as e:
            raise VardefClientError(
                json.dumps(
                    {
                        "status": e.status,
                        "title": "Unauthorized",
                        "detail": "Unauthorized",
                    },
                ),
            ) from e
        except OpenApiException as e:
            raise VardefClientError(e.body) from e

    return _impl


class VariableNotFoundError(Exception):
    """Custom exception for when a variable is not found.

    Attributes:
        message (str): Message describing the error.
    """

    def __init__(self, message: str) -> None:
        """Initialize the VariableNotFoundError.

        Args:
            message (str): Description of the error.
        """
        super().__init__(message)
        self.message = message

    def __str__(self) -> str:
        """Return the string representation of the exception."""
        return f" {self.message}"


class VardefFileError(Exception):
    """Custom exception for catching errors related to variable definition file handling.

    Attributes:
        message (str): Message describing the error.
    """

    def __init__(self, message: str, *args) -> None:  # noqa: ANN002
        """Accepting the message and any additional arguments."""
        super().__init__(message, *args)
        self.message = message
        self.args = args

    def __str__(self) -> str:
        """Returning a custom string representation of the exception."""
        return f"VardefFileError: {self.message}"


def vardef_file_error_handler(method):  # noqa: ANN201, ANN001
    """Decorator for handling exceptions when generating yaml files for variable definitions."""

    @wraps(method)
    def _impl(*method_args, **method_kwargs):  # noqa: ANN002, ANN003
        try:
            return method(
                *method_args,
                **method_kwargs,
            )
        except FileExistsError as e:
            msg = f"File already exists and can not be saved: {method_kwargs.get('file_path', 'unknown file path')}"
            raise VardefFileError(msg) from e
        except PermissionError as e:
            msg = f"Permission denied for file path when accessing the file. Original error: {e!s}"
            raise VardefFileError(msg) from e
        except UnknownTimeZoneError as e:
            msg = f"Timezone is unknown: {method_kwargs.get('time_zone', 'unknown')}"
            raise VardefFileError(msg) from e
        except YAMLError as e:
            msg = f"Invalid yaml. Please fix the formatting in your yaml file.\nOriginal error:\n{e!s}"
            raise VardefFileError(msg) from e
        except EOFError as e:
            msg = "Unexpected end of file"
            raise VardefFileError(msg) from e
        except NotADirectoryError as e:
            msg = f"Path is not a directory: {method_kwargs.get('file_path', 'unknown file path')}. Original error: {e!s}"
            raise VardefFileError(msg) from e

    return _impl


class PublishingBlockedError(RuntimeError):
    """Exception raised when publishing variable definitions is blocked in the production environment."""

    default_message = PUBLISHING_BLOCKED_ERROR_MESSAGE

    def __str__(self) -> str:
        """Return the default publishing blocked error message."""
        return self.default_message


def publishing_blocked_error_handler(method):  # noqa: ANN201, ANN001
    """Decorator that blocks publishing variable definitions in production.

    - If the environment is production:
        - If `variable_status` is present in the arguments and set to `PUBLISHED_INTERNAL` or `PUBLISHED_EXTERNAL`,
        publishing is blocked by raising a `PublishingBlockedError`.
        - If `variable_status` is set to `DRAFT`, the method is allowed to proceed.
        - If no arguments are provided, all publishing attempts are blocked.
    """

    @wraps(method)
    def _impl(*method_args, **method_kwargs):  # noqa: ANN002, ANN003
        try:
            client = VardefClient()
            host = client.get_config().host
        except VardefClientError as e:
            msg = f"Failed to get VardefClient config: {e}"
            raise RuntimeError(msg) from e

        if host == VARDEF_PROD_URL:
            if len(method_args) > 1:
                status = method_args[1].variable_status
                if status == VariableStatus.DRAFT:
                    return method(
                        *method_args,
                        **method_kwargs,
                    )
                if status in (
                    VariableStatus.PUBLISHED_INTERNAL,
                    VariableStatus.PUBLISHED_EXTERNAL,
                ):
                    raise PublishingBlockedError
            raise PublishingBlockedError
        return method(
            *method_args,
            **method_kwargs,
        )

    return _impl



"""Tools and clients for working with the Dapla Metadata system."""

import warnings

warnings.filterwarnings(
    "ignore",
    message="As the c extension couldn't be imported, `google-crc32c` is using a pure python implementation that is significantly slower.",
)

import datadoc_model.all_optional.model as datadoc_model

from . import dapla
from . import datasets
from . import standards
from . import variable_definitions
